# Copyright (C) 2008-2010 Adam Olsen
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.

from gi.repository import Gtk

from xl import xdg
from xl.nls import gettext as _
from xlgui.preferences import widgets
from xlgui import tray

name = _('Appearance')
icon = 'preferences-desktop-theme'
ui = xdg.get_data_path('ui', 'preferences', 'appearance.ui')


class ShowInfoAreaPreference(widgets.CheckPreference):
    default = True
    name = 'gui/show_info_area'


class ShowInfoAreaCoversPreference(widgets.CheckPreference):
    default = True
    name = 'gui/show_info_area_covers'


class SplashPreference(widgets.CheckPreference):
    default = True
    name = 'gui/use_splash'


class ShowTabBarPreference(widgets.CheckPreference):
    default = True
    name = 'gui/show_tabbar'


def _get_system_default_font():
    return Gtk.Widget.get_default_style().font_desc.to_string()


class PlaylistFontPreference(widgets.FontButtonPreference):
    default = _get_system_default_font()
    name = 'gui/playlist_font'


class PlaylistFontResetButtonPreference(widgets.FontResetButtonPreference):
    default = _get_system_default_font()
    name = 'gui/playlist_font_reset_button'
    condition_preference_name = 'gui/playlist_font'


class GtkDarkThemePreference(widgets.CheckPreference):
    default = False
    name = 'gui/gtk_dark_hint'


class UseAlphaTransparencyPreference(widgets.CheckPreference):
    default = False
    name = 'gui/use_alpha'
    restart_required = True


class TransparencyPreference(widgets.ScalePreference, widgets.CheckConditional):
    default = 0.3
    name = 'gui/transparency'
    condition_preference_name = 'gui/use_alpha'

    def __init__(self, preferences, widget):
        widgets.ScalePreference.__init__(self, preferences, widget)
        widgets.CheckConditional.__init__(self)


class TrackCountsPreference(widgets.CheckPreference):
    default = True
    name = 'gui/display_track_counts'

    def apply(self, value=None):
        return_value = widgets.CheckPreference.apply(self, value)

        import xlgui

        xlgui.get_controller().get_panel('collection').load_tree()

        return return_value


class UseTrayPreference(widgets.CheckPreference, widgets.Conditional):
    default = False
    name = 'gui/use_tray'

    def __init__(self, preferences, widget):
        widgets.CheckPreference.__init__(self, preferences, widget)
        widgets.Conditional.__init__(self)
        if not tray.is_supported():
            self.widget.set_tooltip_text(
                _("Tray icons are not supported on your platform")
            )

    def on_check_condition(self):
        return tray.is_supported()


class MinimizeToTrayPreference(widgets.CheckPreference, widgets.CheckConditional):
    default = False
    name = 'gui/minimize_to_tray'
    condition_preference_name = 'gui/use_tray'

    def __init__(self, preferences, widget):
        widgets.CheckPreference.__init__(self, preferences, widget)
        widgets.CheckConditional.__init__(self)


class CloseToTrayPreference(widgets.CheckPreference, widgets.CheckConditional):
    default = False
    name = 'gui/close_to_tray'
    condition_preference_name = 'gui/use_tray'

    def __init__(self, preferences, widget):
        widgets.CheckPreference.__init__(self, preferences, widget)
        widgets.CheckConditional.__init__(self)


class EnsureVisiblePreference(widgets.CheckPreference):
    default = True
    name = 'gui/ensure_visible'


class TabPlacementPreference(widgets.ComboPreference):
    default = 'top'
    name = 'gui/tab_placement'

    def __init__(self, preferences, widget):
        widgets.ComboPreference.__init__(self, preferences, widget)


class ShowStatusBar(widgets.CheckPreference):
    default = True
    name = 'gui/show_status_bar'


class ShowStatusBarCollectionCount(widgets.CheckPreference, widgets.CheckConditional):
    default = True
    name = 'gui/show_status_bar_collection_count'
    condition_preference_name = 'gui/show_status_bar'

    def __init__(self, preferences, widget):
        widgets.CheckPreference.__init__(self, preferences, widget)
        widgets.CheckConditional.__init__(self)


class ShowStatusBarCountTracksInPlaylist(
    widgets.CheckPreference, widgets.CheckConditional
):
    default = True
    name = 'gui/show_status_bar_count_tracks_in_playlist'
    condition_preference_name = 'gui/show_status_bar'

    def __init__(self, preferences, widget):
        widgets.CheckPreference.__init__(self, preferences, widget)
        widgets.CheckConditional.__init__(self)


class ShowStatusBarTimeInPlaylist(widgets.CheckPreference, widgets.CheckConditional):
    default = True
    name = 'gui/show_status_bar_time_in_playlist'
    condition_preference_name = 'gui/show_status_bar'

    def __init__(self, preferences, widget):
        widgets.CheckPreference.__init__(self, preferences, widget)
        widgets.CheckConditional.__init__(self)


"""
class ProgressBarTextFormatPreference(widgets.ComboEntryPreference):
    name = 'gui/progress_bar_text_format'
    completion_items = {
        '$current_time': _('Current playback position'),
        '$remaining_time': _('Remaining playback time'),
        '$total_time': _('Length of a track')
    }
    preset_items = [
        '$current_time / $remaining_time',
        '$current_time / $total_time'
    ]
    default = '$current_time / $remaining_time'
"""

# vim: et sts=4 sw=4



import logging

from gi.repository import GLib

from xl import player, event
from xl import settings as xl_settings
from . import acprefs

LOGGER = logging.getLogger(__name__)


class AlarmClock:
    def __init__(self):
        self.__timeout_id = None
        self.__fade_id = None
        self.__current_volume = None
        event.add_callback(self.__on_option_set, 'plugin_alarmclock_option_set')
        self.__update_timeout()

    def stop_timeout(self):
        if self.__timeout_id is not None:
            GLib.Source.remove(self.__timeout_id)
            self.__timeout_id = None

    def abort_fade(self):
        if self.__fade_id is not None:
            GLib.Source.remove(self.__fade_id)
            self.__fade_id = None
            max_volume = self.__get_pref(acprefs.MaxVolumePreference)
            player.PLAYER.set_volume(max_volume)

    @staticmethod
    def __is_weekday_enabled(day_nr):
        "day_nr starts with 1=Monday, 2=Tuesday, ..., 7=Sunday"
        if day_nr < 1 or day_nr > 7:
            raise ValueError

        weekdays = [
            acprefs.MondayPreference,
            acprefs.TuesdayPreference,
            acprefs.WednesdayPreference,
            acprefs.ThursdayPreference,
            acprefs.FridayPreference,
            acprefs.SaturdayPreference,
            acprefs.SundayPreference,
        ]
        pref_class = weekdays[day_nr - 1]
        return AlarmClock.__get_pref(pref_class)

    @staticmethod
    def __get_pref(pref_class):
        return xl_settings.get_option(pref_class.name, pref_class.default)

    def __update_timeout(self):
        LOGGER.debug("Updating timeout")
        self.stop_timeout()
        current = GLib.DateTime.new_now_local()
        hour = self.__get_pref(acprefs.HourPreference)
        minute = self.__get_pref(acprefs.MinutesPreference)
        tmp_timer = GLib.DateTime.new_local(
            current.get_year(),
            current.get_month(),
            current.get_day_of_month(),
            hour,
            minute,
            0,
        )

        next_timer = None
        for delta_days in range(0, 8):  # current weekday is checked twice intentional
            next_timer = tmp_timer.add_days(delta_days)
            if next_timer.difference(current) <= 0:
                next_timer = None
                continue
            if self.__is_weekday_enabled(next_timer.get_day_of_week()):
                break
            else:
                next_timer = None
                continue

        if next_timer is None:  # no alarm enabled
            LOGGER.debug("No alarm enabled")
            return

        diff = next_timer.difference(current) // 1000
        LOGGER.debug(
            "Alarm due in %i seconds (%i hours)", diff // 1000, diff // 3600000
        )
        self.__timeout_id = GLib.timeout_add(diff, self.__on_timeout_finished)

    def __on_timeout_finished(self):
        LOGGER.debug("Finished timeout")
        # make alarm
        if self.__get_pref(acprefs.FadingPreference):
            min_volume = self.__get_pref(acprefs.MinVolumePreference)
            self.__current_volume = min_volume
            player.PLAYER.set_volume(self.__current_volume)
            timer_step = self.__get_pref(acprefs.TimerStepPreference)
            self.__fade_id = GLib.timeout_add(timer_step * 1000, self.__fade_in)
        else:
            max_volume = self.__get_pref(acprefs.MaxVolumePreference)
            player.PLAYER.set_volume(max_volume)

        if not player.PLAYER.is_playing():
            if player.PLAYER.is_paused():
                player.PLAYER.unpause()
            elif player.PLAYER.current is not None:
                player.PLAYER.play(player.PLAYER.current)
            else:
                # TODO: handle this case better
                LOGGER.error("No song found to play")

        # prepare for next run
        self.__update_timeout()
        return GLib.SOURCE_REMOVE

    def __fade_in(self):
        increment = self.__get_pref(acprefs.IncrementPreference)
        # try to detect whether the user interacted with exaile and abort
        delta_volume = abs(player.PLAYER.get_volume() - self.__current_volume)
        if (
            player.PLAYER.is_paused()
            or not player.PLAYER.is_playing()
            or delta_volume > (increment / 2)
        ):
            self.__fade_id = None
            LOGGER.debug("Aborting fade in because of user interaction")
            return GLib.SOURCE_REMOVE

        new_volume = self.__current_volume + increment

        max_volume = self.__get_pref(acprefs.MaxVolumePreference)
        # check whether we finished fading in
        if new_volume >= max_volume:
            player.PLAYER.set_volume(max_volume)
            self.__fade_id = None
            LOGGER.debug("Finished fading in.")
            return GLib.SOURCE_REMOVE

        self.__current_volume = new_volume
        player.PLAYER.set_volume(new_volume)
        return GLib.SOURCE_CONTINUE

    def __on_option_set(self, _event, _settings, option):
        if 'alarm_' not in option:
            self.__update_timeout()


class AlarmclockPlugin:
    def __init__(self):
        self.__alarm_clock = None

    def enable(self, _exaile):
        self.__alarm_clock = AlarmClock()

    def disable(self, _exaile):
        self.__alarm_clock.stop_timeout()
        self.__alarm_clock.abort_fade()

    def get_preferences_pane(self):
        return acprefs


plugin_class = AlarmclockPlugin



# Copyright (C) 2008-2010 Adam Olsen
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.

"""
Collection of useful stock MenuItems for use with xlgui.widgets.menu
"""

# TODO: how should we document standardization of context's
# selected-(items|tracks) ?

from gi.repository import Gio
from gi.repository import GLib
from gi.repository import Gtk


from xl import common, player, playlist, settings, trax
from xl.nls import gettext as _
from xlgui.widgets import dialogs, rating, menu
from xlgui import panel, properties

### TRACKS ITEMS ###
# These items act on a set of Tracks, by default 'selected-tracks' from
# the parent's context, but custom accessors are allowed via the
# get_tracks_func kwarg


def generic_get_playlist_func(parent, context):
    return context.get('selected-playlist', None)


def generic_get_tracks_func(parent, context):
    return context.get('selected-tracks', [])


class RatingMenuItem(menu.MenuItem):
    """
    A menu item displaying rating images
    and allowing for selection of ratings
    """

    def __init__(self, name, after, get_tracks_func=generic_get_tracks_func):
        menu.MenuItem.__init__(self, name, self.factory, after)
        self.get_tracks_func = get_tracks_func
        self.rating_set = False

        # TODO: For accessibility it would be nice to add mnemonics or some
        # other key shortcut thing to the RatingMenu, e.g. "+" and "-"

    def factory(self, menu, parent, context):
        # don't show rating widget for computed track selections (see #340)
        if context.get('needs-computing'):
            return
        item = rating.RatingMenuItem()
        item.connect('show', self.on_show, menu, parent, context)
        self._rating_changed_id = item.connect(
            'rating-changed', self.on_rating_changed, menu, parent, context
        )

        return item

    # @common.threaded
    # TODO: ASYNC
    # most of this function isn't safe to use from a (interacts with UI elements)
    def on_show(self, widget, menu, parent, context):
        """
        Updates the menu item on show
        """
        tracks = self.get_tracks_func(parent, context)
        rating = trax.util.get_rating_from_tracks(tracks)
        widget.disconnect(self._rating_changed_id)
        widget.props.rating = rating
        self._rating_changed_id = widget.connect(
            'rating-changed', self.on_rating_changed, menu, parent, context
        )

    def on_rating_changed(self, widget, rating, menu, parent, context):
        """
        Passes the 'rating-changed' signal
        """
        tracks = self.get_tracks_func(parent, context)
        for track in tracks:
            track.set_rating(rating)


def _enqueue_cb(widget, name, parent, context, get_tracks_func):
    tracks = get_tracks_func(parent, context)
    player.QUEUE.extend(tracks)


def EnqueueMenuItem(name, after, get_tracks_func=generic_get_tracks_func):
    return menu.simple_menu_item(
        name,
        after,
        _("En_queue"),
        'list-add',
        _enqueue_cb,
        callback_args=[get_tracks_func],
    )


# TODO: move logic into (GUI?) playlist


def _append_cb(widget, name, parent, context, get_tracks_func, replace=False):
    from xlgui import main

    page = main.get_selected_playlist()
    if not page:
        return
    pl = page.playlist
    if replace:
        pl.clear()
    offset = len(pl)
    tracks = get_tracks_func(parent, context)
    # This is well-intentioned, but it leads to odd effects (see #147)
    # -> this is more proof that the playlist needs to handle this logic
    # sort_by, reverse = page.view.get_sort_by()
    # tracks = trax.sort_tracks(sort_by, tracks, reverse=reverse,
    #    artist_compilations=True)
    pl.extend(tracks)
    if settings.get_option('playlist/append_menu_starts_playback', False):
        if not player.PLAYER.current:
            page.view.play_track_at(offset, tracks[0])


def ReplaceCurrentMenuItem(name, after, get_tracks_func=generic_get_tracks_func):
    return menu.simple_menu_item(
        name,
        after,
        _("_Replace Current"),
        None,
        _append_cb,
        callback_args=[get_tracks_func, True],
    )


def AppendMenuItem(name, after, get_tracks_func=generic_get_tracks_func):
    return menu.simple_menu_item(
        name,
        after,
        _("_Append to Current"),
        'list-add',
        _append_cb,
        callback_args=[get_tracks_func],
    )


def _properties_cb(widget, name, parent, context, get_tracks_func):
    tracks = get_tracks_func(parent, context)
    if tracks:
        assert isinstance(parent, panel.Panel)
        properties.TrackPropertiesDialog(parent.parent, tracks)


def PropertiesMenuItem(name, after, get_tracks_func=generic_get_tracks_func):
    return menu.simple_menu_item(
        name,
        after,
        _("_Track Properties"),
        'document-properties',
        _properties_cb,
        callback_args=[get_tracks_func],
    )


def _open_directory_cb(widget, name, parent, context, get_tracks_func):
    try:
        track = get_tracks_func(parent, context)[0]
    except IndexError:
        return
    common.open_file_directory(track.get_loc_for_io())


def OpenDirectoryMenuItem(name, after, get_tracks_func=generic_get_tracks_func):
    return menu.simple_menu_item(
        name,
        after,
        _("_Open Directory"),
        'folder-open',
        _open_directory_cb,
        callback_args=[get_tracks_func],
    )


def generic_trash_tracks_func(parent, context, tracks):
    for track in tracks:
        gfile = Gio.File.new_for_uri(track.get_loc_for_io())
        gfile.trash()


def generic_delete_tracks_func(parent, context, tracks):
    for track in tracks:
        gfile = Gio.File.new_for_uri(track.get_loc_for_io())
        gfile.delete()


def _on_trash_tracks(
    widget,
    name,
    parent,
    context,
    get_tracks_func,
    trash_tracks_func,
    delete_tracks_func,
):
    tracks = get_tracks_func(parent, context)

    try:
        trash_tracks_func(parent, context, tracks)
    except GLib.Error:
        dialog = Gtk.MessageDialog(
            parent=parent.parent,
            message_type=Gtk.MessageType.WARNING,
            text=_(
                'The files cannot be moved to the Trash. '
                'Delete them permanently from the disk?'
            ),
        )
        dialog.add_buttons(
            Gtk.STOCK_CANCEL,
            Gtk.ResponseType.CANCEL,
            Gtk.STOCK_DELETE,
            Gtk.ResponseType.OK,
        )
        dialog.set_alternative_button_order_from_array(
            (Gtk.ResponseType.OK, Gtk.ResponseType.CANCEL)
        )

        if dialog.run() == Gtk.ResponseType.OK:
            delete_tracks_func(parent, context, tracks)

        dialog.destroy()


def TrashMenuItem(
    name,
    after,
    get_tracks_func=generic_get_tracks_func,
    trash_tracks_func=generic_trash_tracks_func,
    delete_tracks_func=generic_delete_tracks_func,
):
    return menu.simple_menu_item(
        name,
        after,
        _('_Move to Trash'),
        'user-trash',
        _on_trash_tracks,
        callback_args=[get_tracks_func, trash_tracks_func, delete_tracks_func],
    )


### END TRACKS ITEMS ###

### PLAYLIST ITEMS ###
# These items act on a playlist, by default 'selected-playlist' from
# the parent's context, but custom accessors are allowed via the
# get_pl_func kwarg


def RenamePlaylistMenuItem(name, after, get_pl_func=generic_get_playlist_func):
    return menu.simple_menu_item(
        name,
        after,
        _('_Rename'),
        'accessories-text-editor',
        lambda w, n, o, c: o.rename_playlist(get_pl_func(o, c)),
        condition_fn=lambda n, p, c: not isinstance(
            c['selected-playlist'], playlist.SmartPlaylist
        ),
    )


def EditPlaylistMenuItem(name, after, get_pl_func=generic_get_playlist_func):
    return menu.simple_menu_item(
        name,
        after,
        _('_Edit'),
        'accessories-text-editor',
        lambda w, n, o, c: o.edit_smart_playlist(get_pl_func(o, c)),
        condition_fn=lambda n, p, c: isinstance(
            c['selected-playlist'], playlist.SmartPlaylist
        ),
    )


def ExportPlaylistMenuItem(name, after, get_pl_func=generic_get_playlist_func):
    return menu.simple_menu_item(
        name,
        after,
        _('E_xport Playlist'),
        'document-save-as',
        lambda w, n, o, c: dialogs.export_playlist_dialog(get_pl_func(o, c)),
    )


def ExportPlaylistFilesMenuItem(name, after, get_pl_func=generic_get_playlist_func):
    return menu.simple_menu_item(
        name,
        after,
        _('Export _Files'),
        'document-save-as',
        lambda w, n, o, c: dialogs.export_playlist_files(get_pl_func(o, c)),
    )


def DeletePlaylistMenuItem(name, after, get_pl_func=generic_get_playlist_func):
    return menu.simple_menu_item(
        name,
        after,
        _('_Delete Playlist'),
        'edit-delete',
        lambda w, n, o, c: o.remove_playlist(get_pl_func(o, c)),
    )


### END PLAYLIST ITEMS ###



# Copyright (C) 2015 Dustin Spicuzza
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

from xl import settings


def migrate():
    """
    Migrate normal/unified engine users to gstreamer
    """
    engine = settings.get_option('player/engine', None)
    if engine in ['normal', 'unified']:
        settings.set_option('player/engine', 'gstreamer')



# Copyright (C) 2008-2010 Adam Olsen
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.

import contextlib
import logging
import os.path
import sys
from typing import Optional, Tuple, Union

from gi.repository import Gio
from gi.repository import Gdk
from gi.repository import GdkPixbuf
from gi.repository import GLib
from gi.repository import Gtk
from gi.repository import Pango

from xl import settings, xdg
from xl.nls import gettext as _

# Required for xlgui.get_controller()
import xlgui

# Import from external namespace
from xl.externals.gi_composites import GtkTemplate as _GtkTemplate

logger = logging.getLogger(__name__)


class GtkTemplate(_GtkTemplate):
    """
    Use this class decorator in conjunction with :class:`.GtkCallback`
    and :class:`GtkChild` to construct widgets from a GtkBuilder UI
    file.

    This is an exaile-specific wrapper around the :class:`.GtkTemplate`
    object to allow loading the UI template file in an Exaile-specific
    way.

    :param *path: Path components to specify UI file
    :param relto: If keyword arg 'relto' is specified, path will be
                  relative to this. Otherwise, it will be relative to
                  the Exaile data directory

    .. versionadded:: 3.5.0
    """

    def __init__(self, *path, **kwargs):
        super(GtkTemplate, self).__init__(ui=ui_path(*path, **kwargs))


def ui_path(*path, **kwargs):
    """
    Returns absolute path to a UI file. Each arg will be concatenated
    to construct the final path.

    :param relto: If keyword arg 'relto' is specified, path will be
                  relative to this. Otherwise, it will be relative to
                  the Exaile data directory

    .. versionadded:: 3.5.0
    """

    relto = kwargs.pop('relto', None)
    if len(kwargs):
        raise ValueError("Only 'relto' is allowed as a keyword argument")

    if relto is None:
        return xdg.get_data_path(*path)
    else:
        return os.path.abspath(os.path.join(os.path.dirname(relto), *path))


def get_workarea_size():
    """
    Returns the width and height of the work area
    """
    d = get_workarea_dimensions()
    return (d.width, d.height)


def get_workarea_dimensions(window=None):
    """
    Returns the x-offset, y-offset, width and height of the work area
    for a given window or for the default screen if no window is given.
    Falls back to the screen dimensions if not available.

    :param window: class: `Gtk.Window`, optional

    :returns: :class:`CairoRectangleInt`
    """
    if window is None:
        screen = Gdk.Screen.get_default()
        default_monitor = screen.get_primary_monitor()
        return screen.get_monitor_workarea(default_monitor)
    else:
        display = window.get_window().get_display()
        work_area = display.get_monitor_at_window(window.get_window()).get_workarea()
    return work_area


def gtk_widget_replace(widget, replacement):
    """
    Replaces one widget with another and places it exactly at the
    original position, keeping child properties

    :param widget: The original widget
    :type widget: :class:`Gtk.Widget`
    :param replacement: The new widget
    :type widget: :class:`Gtk.Widget`

    :returns: replacement widget if successful
    """
    parent = widget.get_parent()

    if parent is None:
        logger.error("widget doesn't have a parent.")
        return

    props = {}
    for pspec in parent.list_child_properties():
        props[pspec.name] = parent.child_get_property(widget, pspec.name)

    parent.remove(widget)
    parent.add(replacement)

    for name, value in props.items():
        parent.child_set_property(replacement, name, value)
    return replacement


def pixbuf_from_data(
    data: bytes,
    size: Optional[Tuple[int, int]] = None,
    keep_ratio: bool = True,
    upscale: bool = False,
) -> Optional[GdkPixbuf.Pixbuf]:
    """
    Generates a pixbuf from arbitrary image data

    :param data: The raw image data
    :param size: Size to scale to; if not specified,
        the image will render to its native size
    :param keep_ratio: Whether to keep the original
        image ratio on resizing operations
    :param upscale: Whether to upscale if the requested
        size exceeds the native size

    :returns: the generated pixbuf
    """
    if not data:
        return None

    pixbuf = None
    loader = GdkPixbuf.PixbufLoader()

    if size is not None:

        def on_size_prepared(loader, width, height):
            """
            Keeps the ratio if requested
            """
            if keep_ratio:
                scale = min(size[0] / float(width), size[1] / float(height))

                if scale > 1.0 and upscale:
                    width = int(width * scale)
                    height = int(height * scale)
                elif scale <= 1.0:
                    width = int(width * scale)
                    height = int(height * scale)
            else:
                if upscale:
                    width, height = size
                else:
                    width = height = max(width, height)

            loader.set_size(width, height)

        loader.connect('size-prepared', on_size_prepared)

    try:
        loader.write(data)
        loader.close()
    except GLib.Error as e:
        logger.warning(
            'Failed to get pixbuf from data: {error}'.format(error=e.message)
        )
    else:
        pixbuf = loader.get_pixbuf()

    return pixbuf


class ScalableImageWidget(Gtk.Image):
    """
    Custom resizable image widget
    """

    def __init__(self):
        """
        Initializes the image
        """
        Gtk.Image.__init__(self)

    def set_image_size(self, width, height):
        """
        Scales the size of the image
        """
        self.size = (width, height)
        self.set_size_request(width, height)

    def set_image(self, location: str, fill: bool = False):
        """
        Sets the image from a location

        :param location: the location to load the image from
        :param fill: True to expand the image, False to keep its ratio
        """
        pixbuf = GdkPixbuf.Pixbuf.new_from_file(
            Gio.File.new_for_uri(location).get_path()
        )
        self.set_image_pixbuf(pixbuf, fill)

    def set_image_data(self, data: bytes, fill: bool = False):
        """
        Sets the image from binary data

        :param data: the binary data
        :param fill: True to expand the image, False to keep its ratio
        """
        if not data:
            return
        pixbuf = pixbuf_from_data(data)
        self.set_image_pixbuf(pixbuf, fill)

    def set_image_pixbuf(self, pixbuf: GdkPixbuf.Pixbuf, fill: bool = False):
        """
        Sets the image from a pixbuf

        :param data: the pixbuf
        :param fill: True to expand the image, False to keep its ratio
        """
        width, height = self.size
        if not fill:
            origw = float(pixbuf.get_width())
            origh = float(pixbuf.get_height())
            scale = min(width / origw, height / origh)
            width = int(origw * scale)
            height = int(origh * scale)
        self.width = width
        self.height = height
        scaled = pixbuf.scale_simple(width, height, GdkPixbuf.InterpType.BILINEAR)
        self.set_from_pixbuf(scaled)


class SearchEntry:
    """
    A Gtk.Entry that emits the "activated" signal when something has
    changed after the specified timeout
    """

    def __init__(self, entry: Optional[Gtk.Entry] = None, timeout: int = 500):
        """
        Initializes the entry

        :param timeout: Timeout in ms
        """
        self.timeout = timeout
        self.change_id = None

        if entry is None:
            entry = Gtk.Entry.new()
        self.entry = entry

        self._last_text = entry.get_text()

        entry.connect('changed', self.on_entry_changed)
        entry.connect('icon-press', self.on_entry_icon_press)
        entry.connect('activate', self.on_entry_activated)

        # Disable global accelerators while editing, re-enable when
        # done
        entry.connect(
            'focus-in-event',
            lambda w, e: xlgui.get_controller().main.accel_manager.disable_accelerators(),
        )
        entry.connect(
            'focus-out-event',
            lambda w, e: xlgui.get_controller().main.accel_manager.enable_accelerators(),
        )

    def on_entry_changed(self, entry: Gtk.Entry):
        """
        Called when the entry changes
        """
        empty_search = entry.get_text() == ''
        entry.props.secondary_icon_sensitive = not empty_search

        if self.change_id:
            GLib.source_remove(self.change_id)
        self.change_id = GLib.timeout_add(self.timeout, self.entry_activate)

    def on_entry_icon_press(self, entry, icon_pos, event):
        """
        Clears the entry
        """
        self.entry.set_text('')
        self.entry_activate()

    def on_entry_activated(self, entry):
        self._last_text = entry.get_text()

    def entry_activate(self, *e):
        """
        Emit the activate signal
        """
        self.change_id = None
        if self.entry.get_text() != self._last_text:
            self.entry.activate()

    def __getattr__(self, attr):
        """
        Tries to pass attribute requests
        to the internal entry item
        """
        return getattr(self.entry, attr)


class ModifierType:
    """
    Common Gdk.ModifierType combinations that work in a cross platform way
    """

    #
    # Missing from Gdk.ModifierType
    #

    #: Apple/cmd on OSX, CTRL elsewhere (taken from QuodLibet)
    PRIMARY_MASK = Gtk.accelerator_parse("<Primary>")[1]

    #: primary + shift
    PRIMARY_SHIFT_MASK = PRIMARY_MASK | Gdk.ModifierType.SHIFT_MASK

    #
    # The rest are for completeness..
    #

    #: shift
    SHIFT_MASK = Gdk.ModifierType.SHIFT_MASK


def position_menu(menu, *args):
    """
    A function that will position a menu near a particular widget. This
    should be specified as the third argument to menu.popup(), with the
    user data being the widget.

        menu.popup_menu(None, None, guiutil.position_menu, widget,
                        0, 0)
    """
    # Prior to GTK+ 3.16, args contains only our user data.
    # Since 3.16, we get (orig_x, orig_y, data).
    # See https://git.gnome.org/browse/gtk+/commit/?id=8463d0ee62b4b22fa
    widget = args[-1]
    window = widget.get_window()
    _, window_x, window_y = window.get_origin()
    widget_allocation = widget.get_allocation()
    menu_allocation = menu.get_allocation()
    position = (
        window_x + widget_allocation.x + 1,
        window_y + widget_allocation.y - menu_allocation.height - 1,
    )

    return (position[0], position[1], True)


def finish(repeat=True):
    """
    Waits for current pending gtk events to finish
    """
    while Gtk.events_pending():
        Gtk.main_iteration()
        if not repeat:
            break


def initialize_from_xml(this, other=None):
    """
    DEPRECATED. Use GtkComposite, GtkCallback, and GtkChild instead

    Initializes the widgets and signals from a GtkBuilder XML file. Looks
    for the following attributes in the instance you pass:

    ui_filename = builder filename -- either an absolute path, or a tuple
                  with the path relative to the xdg data directory.
    ui_widgets = [list of widget names]
    ui_signals = [list of function names to connect to a signal]

    For each widget in ui_widgets, it will be retrieved from the builder
    object and set as an attribute on the object you pass in.

    other is a list of widgets to also initialize with the same file

    Returns the builder object when done
    """
    builder = Gtk.Builder()

    if isinstance(this.ui_filename, str):
        builder.add_from_file(this.ui_filename)
    else:
        builder.add_from_file(xdg.get_data_path(*this.ui_filename))

    objects = [this]
    if other is not None:
        objects.extend(other)

    for obj in objects:
        if hasattr(obj, 'ui_widgets') and obj.ui_widgets is not None:
            for widget_name in obj.ui_widgets:
                widget = builder.get_object(widget_name)
                if widget is None:
                    raise RuntimeError(
                        "Widget '%s' is not present in '%s'"
                        % (widget_name, this.ui_filename)
                    )
                setattr(obj, widget_name, widget)

    signals = None

    for obj in objects:
        if hasattr(obj, 'ui_signals') and obj.ui_signals is not None:
            if signals is None:
                signals = {}
            for signal_name in obj.ui_signals:
                if not hasattr(obj, signal_name):
                    raise RuntimeError(
                        "Function '%s' is not present in '%s'" % (signal_name, obj)
                    )
                signals[signal_name] = getattr(obj, signal_name)

    if signals is not None:
        missing = builder.connect_signals(signals)
        if missing is not None:
            err = (
                'The following signals were found in %s but have no assigned handler: %s'
                % (this.ui_filename, str(missing))
            )
            raise RuntimeError(err)

    return builder


def persist_selection(widget, key_col, setting_name):
    """
    Given a widget that is using a Gtk.ListStore, it will restore the
    selected index given the contents of a setting. When the widget
    changes, it will save the choice.

    Call this on the widget after you have loaded data
    into the widget.

    :param widget:         Gtk.ComboBox or Gtk.TreeView
    :param col:            Integer column with unique key
    :param setting_name:   Setting to save key to/from
    """

    model = widget.get_model()

    key = settings.get_option(setting_name)
    if key is not None:
        for i in range(len(model)):
            if model[i][key_col] == key:
                if hasattr(widget, 'set_active'):
                    widget.set_active(i)
                else:
                    widget.set_cursor((i,))
                break

    if hasattr(widget, 'set_active'):

        def _on_changed(widget):
            active = widget.get_model()[widget.get_active()][key_col]
            settings.set_option(setting_name, active)

        widget.connect('changed', _on_changed)

    else:

        def _on_changed(widget):
            model, i = widget.get_selected()
            active = model[i][key_col]
            settings.set_option(setting_name, active)

        widget.get_selection().connect('changed', _on_changed)


def platform_is_wayland():
    """
    This function checks whether Exaile has been started on a Wayland display.

    This function has been tested on both GNOME and Weston Wayland compositors

    :returns: `True` if the display used by Exaile is using a Wayland compositor,
                ` False` otherwise.
    """
    display_name = Gdk.Display.get_default().get_name().lower()
    return 'wayland' in display_name


def platform_is_x11():
    """
    This function checks whether Exaile has been started on a X11 Gdk backend.

    :returns: `True` if the display used by Exaile is realized on a X11 server,
                ` False` otherwise.
    """
    display_name = Gdk.Display.get_default().get_name().lower()
    return 'x11' in display_name


def css_from_rgba(rgba):
    """
    Convert a Gdk.RGBA to a CSS color string
    """
    color_css_str = "rgba(%s, %s, %s, %s)" % (
        str(int(rgba.red * 255)),
        str(int(rgba.green * 255)),
        str(int(rgba.blue * 255)),
        str(rgba.alpha),
    )
    return color_css_str


def css_from_rgba_without_alpha(rgba):
    """
    Convert a Gdk.RGBA to a CSS color string removing the alpha channel
    """
    color_css_str = "rgb(%s, %s, %s)" % (
        str(int(rgba.red * 255)),
        str(int(rgba.green * 255)),
        str(int(rgba.blue * 255)),
    )
    return color_css_str


def css_from_pango_font_description(pango_font_str):
    """
    Convert a Pango.FontDescription string to a CSS font string
    """
    if pango_font_str is None:
        return ""

    new_font = Pango.FontDescription.from_string(pango_font_str)

    # Gtk+ CSS and Pango are compatible except for the prefix and case
    style_name = new_font.get_style().value_name
    style = style_name.split('_', 2)[2].lower()

    # Gtk+ CSS only allows 100 | 200 | 300 | 400 | 500 | 600 | 700 | 800 | 900
    # Pango allows everything from 100 through 1000, with several named values
    # at least the numbers mean the same
    # Bug in PyGObject: This does not work
    weight_int = int(new_font.get_weight())
    css_weight_int = ((weight_int + 50) // 100) * 100
    if css_weight_int > 900:
        css_weight_int = 900
    weight = str(css_weight_int)

    # Gtk+ CSS and Pango are compatible except for
    # * the prefix
    # * letter case
    # * the fact that Pango uses an underscore whereas Gtk+ CSS uses a dash
    stretch_name = new_font.get_stretch().value_name
    stretch = stretch_name.split('_', 2)[2].lower().replace('_', '-')

    # Gtk+ CSS and Pango are compatible except for
    # * the prefix
    # * letter case
    # * the fact that Pango uses an underscore whereas Gtk+ CSS uses a dash
    variant_name = new_font.get_variant().value_name
    variant = variant_name.split('_', 2)[2].lower().replace('_', '-')

    # Pango multiplies its font by Pango.SCALE
    size = str(new_font.get_size() / Pango.SCALE) + "pt"

    # See "GTK+ CSS" documentation page for the syntax

    # According to https://www.w3schools.com/cssref/pr_font_font-family.asp
    # "If a font name contains white-space, it must be quoted"
    font_css_str = 'font: %s %s %s %s %s "%s"' % (
        style,
        variant,
        weight,
        stretch,
        size,
        new_font.get_family(),
    )
    return font_css_str


def _get_closest_visible(model, child_path):
    # returns closest visible path in filter model to the given child_path
    # (which is presumed to not be visible in the model)
    # -> derived from python's bisect_left (Python license)
    x = child_path.get_indices()[0]

    ll = len(model)
    lo = 0
    hi = ll
    if not hi:
        return

    while lo < hi:
        mid = (lo + hi) // 2
        amid = model.convert_path_to_child_path(model[mid].path).get_indices()[0]
        if amid < x:
            lo = mid + 1
        else:
            hi = mid
    if lo == ll:
        lo -= 1
    return model[lo].path


@contextlib.contextmanager
def without_model(tv):
    """
    Context manager that removes the model from a treeview and restores the
    view position and selection when finished.

    Issue #199: Anytime a large number of changes are made to the model,
    it's better to just remove it before making the changes, as the UI
    will freeze if you don't. However, that messes up the visible view, so
    this saves/restores the view information.

    .. note:: Assumes the model is a Gtk.TreeModelFilter
    """
    model = tv.get_model()
    child_model = model.get_model()
    selection = tv.get_selection()

    # Save existing scroll information if this is a modelfilter
    visible_data = tv.get_visible_range()
    if visible_data:
        # save a ref and path data -- path is used if the row disappears
        visible_data = model.convert_path_to_child_path(visible_data[0])
        visible_ref = Gtk.TreeRowReference(child_model, visible_data)

    # grab selection data, convert them to child row references
    # -> need refs because paths will change if objects are deleted
    seldata = selection.get_selected_rows()
    selected_rows = []
    first_selected_path = None
    if seldata:
        for path in seldata[1]:
            path = model.convert_path_to_child_path(path)
            if not first_selected_path:
                first_selected_path = path
            selected_rows.append(Gtk.TreeRowReference(child_model, path))

    tv.set_model(None)
    yield model
    tv.set_model(model)

    # restore selection data first
    for row in selected_rows:
        if not row.valid():
            continue
        path = model.convert_child_path_to_path(row.get_path())
        if path:
            selection.select_path(path)
            first_selected_path = None

    # If there were no visible iters, try to find something close
    if first_selected_path:
        first_selected_path = _get_closest_visible(model, first_selected_path)
        if first_selected_path:
            selection.select_path(first_selected_path)

    # Restore scroll position
    if visible_data:
        # If original item is visible, this is easy
        if visible_ref.valid():
            # Use the ref first if it's still valid
            scroll_to = model.convert_child_path_to_path(visible_ref.get_path())
        else:
            # But if it isn't valid, then just use the path data
            scroll_to = model.convert_child_path_to_path(visible_data)

        if scroll_to is None:
            # If not, search for the closest visible item
            scroll_to = _get_closest_visible(model, visible_data)

        if scroll_to:
            tv.scroll_to_cell(scroll_to, None, True, 0, 0)


def get_builder(uifile: str) -> Gtk.Builder:
    """
    Loads Gtk.Builder with uifile.
    In case of platform win32 each translatable node in the ui is translated here

    Relates to https://gitlab.gnome.org/GNOME/pygobject/-/issues/422

    :param uifile: Path to uifile
    """
    builder = Gtk.Builder()

    if sys.platform == 'win32':
        with open(uifile, 'rb') as fp:
            template = fp.read()
            template_string = get_template_translated(template)
            builder.add_from_string(template_string)

    else:
        builder.add_from_file(uifile)

    return builder


def get_template_translated(template_ui_xml: Union[str, bytes]) -> str:
    """
    Insert translations into template_ui_xml for each translatable element
    """
    import xml.etree.ElementTree as ET

    tree = ET.fromstring(template_ui_xml)
    for node in tree.iter():
        if 'translatable' in node.attrib:
            node.text = _(node.text)
    xml_text = ET.tostring(tree, encoding='unicode', method='xml')
    return xml_text


# vim: et sts=4 sw=4



# This plugin is adapted from the Python Console plugin and the IPython
# cookbook at:
#   http://ipython.scipy.org/moin/Cookbook/EmbeddingInGTK
# Copyright (C) 2009-2010 Brian Parma
# Updated       2012 Brian Parma
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.


import logging
import sys
import site

from gi.repository import Gdk
from gi.repository import Gtk
from gi.repository import GLib

from xl.nls import gettext as _
from xl import event
from xl import settings as xl_settings
from xl import providers
from xlgui.widgets import menu
from xlgui import guiutil

from . import ipconsoleprefs
from . import ipython_view as ip

FONT = "Luxi Mono 10"
SETTINGS_STRING = 'plugin_ipconsole_option_set'
LOGGER = logging.getLogger(__name__)


class Quitter:
    """Simple class to handle exit, similar to Python 2.5's.

    This Quitter is used to circumvent IPython's circumvention
    of the builtin Quitter, since it prevents exaile form closing."""

    def __init__(self, exit_function, name):
        self.exit_function = exit_function
        self.name = name

    def __repr__(self):
        return 'Type %s() to exit.' % self.name

    def __call__(self):
        self.exit_function()  # Passed in exit function
        site.setquit()  # Restore default builtins
        exit()  # Call builtin


class IPView(ip.IPythonView):
    '''Extend IPythonView to support closing with Ctrl+D'''

    __text_color = None
    __background_color = None
    __font = None

    __css_provider = None

    __text_color_str = None
    __background_color_str = None
    __font_str = None

    __iptheme = None

    def __init__(self, namespace):
        ip.IPythonView.__init__(self)
        event.add_ui_callback(self.__on_option_set, SETTINGS_STRING)
        self.set_wrap_mode(Gtk.WrapMode.CHAR)

        self.updateNamespace(namespace)  # expose exaile (passed in)

        # prevent exit and quit - freezes window? does bad things
        self.updateNamespace({'exit': None, 'quit': None})

        style_context = self.get_style_context()
        self.__css_provider = Gtk.CssProvider()
        style_context.add_provider(
            self.__css_provider, Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION
        )
        # Trigger setup through options
        for option in ('text_color', 'background_color', 'font'):
            self.__on_option_set(
                None, xl_settings, 'plugin/ipconsole/{option}'.format(option=option)
            )

    def __on_option_set(self, _event, settings, option):
        if option == 'plugin/ipconsole/font':
            pango_font_str = settings.get_option(option, FONT)
            self.__font_str = guiutil.css_from_pango_font_description(pango_font_str)
            GLib.idle_add(self.__update_css)
        if option == 'plugin/ipconsole/text_color':
            rgba_str = settings.get_option(option, 'lavender')
            rgba = Gdk.RGBA()
            rgba.parse(rgba_str)
            self.__text_color_str = "color: " + guiutil.css_from_rgba_without_alpha(
                rgba
            )
            GLib.idle_add(self.__update_css)
        if option == 'plugin/ipconsole/background_color':
            rgba_str = settings.get_option(option, 'black')
            rgba = Gdk.RGBA()
            rgba.parse(rgba_str)
            self.__background_color_str = (
                "background-color: " + guiutil.css_from_rgba_without_alpha(rgba)
            )
            GLib.idle_add(self.__update_css)

    def __update_css(self):
        if (
            self.__text_color_str is None
            or self.__background_color_str is None
            or self.__font_str is None
        ):
            # early initialization state: not all properties have been initialized yet
            return False

        data_str = "text {%s; %s;} textview {%s;}" % (
            self.__background_color_str,
            self.__text_color_str,
            self.__font_str,
        )
        self.__css_provider.load_from_data(data_str.encode('utf-8'))
        return False

    def onKeyPressExtend(self, key_event):
        if ip.IPythonView.onKeyPressExtend(self, key_event):
            return True
        if key_event.string == '\x04':  # ctrl+d
            self.destroy()


class IPythonConsoleWindow(Gtk.Window):
    """
    A Gtk Window with an embedded IPython Console.
    """

    __ipv = None

    def __init__(self, namespace):
        Gtk.Window.__init__(self)
        self.set_title(_("IPython Console - Exaile"))
        self.set_size_request(750, 550)
        self.set_resizable(True)

        self.__ipv = IPView(namespace)
        self.__ipv.connect('destroy', lambda *_widget: self.destroy())
        self.__ipv.updateNamespace({'self': self})  # Expose self to IPython

        # make it scrollable
        scrolled_window = Gtk.ScrolledWindow()
        scrolled_window.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
        scrolled_window.add(self.__ipv)
        scrolled_window.show_all()
        self.add(scrolled_window)

        event.add_ui_callback(self.on_option_set, SETTINGS_STRING)

    def on_option_set(self, _event, settings, option):
        if option == 'plugin/ipconsole/opacity':
            if sys.platform.startswith("win32"):
                # Setting opacity on Windows crashes with segfault,
                # see https://bugzilla.gnome.org/show_bug.cgi?id=674449
                # Ignore this option.
                return
            value = settings.get_option(option, 80.0)
            value = value / 100
            if value > 1:
                value = 1
            self.set_opacity(value)


class IPConsolePlugin:
    """
    This class holds the IPConsole plugin itself
    """

    __console_window = None
    __exaile = None

    def enable(self, exaile):
        """
        Called when plugin is enabled, or when exaile is loaded with the plugin
        on by default.
        """
        self.__exaile = exaile

    def on_gui_loaded(self):
        """
        Called when Exaile finished loading its GUI
        """
        # Trigger initial setup through options:
        if xl_settings.get_option('plugin/ipconsole/autostart', False):
            self.__show_console()

        # add menuitem to tools menu
        item = menu.simple_menu_item(
            'ipconsole',
            ['plugin-sep'],
            _('Show _IPython Console'),
            callback=lambda *_args: self.__show_console(),
        )
        providers.register('menubar-tools-menu', item)

    def teardown(self, _exaile):
        """
        Called when Exaile is shutting down
        """
        # if window is open, kill it
        if self.__console_window is not None:
            self.__console_window.destroy()

    def disable(self, exaile):
        """
        Called when the plugin is disabled
        """
        for item in providers.get('menubar-tools-menu'):
            if item.name == 'ipconsole':
                providers.unregister('menubar-tools-menu', item)
                break
        self.teardown(exaile)

    def __show_console(self):
        """
        Display window when the menu item is clicked.
        """
        if self.__console_window is None:
            import xl
            import xlgui

            self.__console_window = IPythonConsoleWindow(
                {'exaile': self.__exaile, 'xl': xl, 'xlgui': xlgui}
            )
            self.__console_window.connect('destroy', self.__console_destroyed)

        self.__console_window.present()
        self.__console_window.on_option_set(
            None, xl_settings, 'plugin/ipconsole/opacity'
        )

    def __console_destroyed(self, *_args):
        """
        Called when the window is closed.
        """
        self.__console_window = None

    def get_preferences_pane(self):
        """
        Called by Exaile when ipconsole preferences pane should be shown
        """
        return ipconsoleprefs


plugin_class = IPConsolePlugin



# Copyright (C) 2008-2010 Adam Olsen
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.

from gi.repository import Gdk
from gi.repository import GdkPixbuf
from gi.repository import Gio
from gi.repository import GLib
from gi.repository import GObject
from gi.repository import Gtk

import xl.radio
import xl.playlist
from xl import event, common, settings, trax
from xl.nls import gettext as _
import xlgui.panel.playlists as playlistpanel
from xlgui.panel import menus
from xlgui import icons, panel
from xlgui.widgets.common import DragTreeView
from xlgui.widgets import dialogs, menu


class RadioException(Exception):
    pass


class ConnectionException(RadioException):
    pass


class RadioPanel(panel.Panel, playlistpanel.BasePlaylistPanelMixin):
    """
    The Radio Panel
    """

    __gsignals__ = {
        'playlist-selected': (GObject.SignalFlags.RUN_LAST, None, (object,)),
        'append-items': (GObject.SignalFlags.RUN_LAST, None, (object, bool)),
        'replace-items': (GObject.SignalFlags.RUN_LAST, None, (object,)),
        'queue-items': (GObject.SignalFlags.RUN_LAST, None, (object,)),
    }
    __gsignals__.update(playlistpanel.BasePlaylistPanelMixin._gsignals_)

    ui_info = ('radio.ui', 'RadioPanel')
    _radiopanel = None

    def __init__(self, parent, collection, radio_manager, station_manager, name):
        """
        Initializes the radio panel
        """
        panel.Panel.__init__(self, parent, name, _('Radio'))
        playlistpanel.BasePlaylistPanelMixin.__init__(self)

        self.collection = collection
        self.manager = radio_manager
        self.playlist_manager = station_manager
        self.nodes = {}
        self.load_nodes = {}
        self.complete_reload = {}
        self.loaded_nodes = []

        self._setup_tree()
        self._setup_widgets()
        self.playlist_image = icons.MANAGER.pixbuf_from_icon_name(
            'music-library', Gtk.IconSize.SMALL_TOOLBAR
        )

        # menus
        self.playlist_menu = menus.RadioPanelPlaylistMenu(self)
        self.track_menu = menus.TrackPanelMenu(self)
        self._connect_events()

        self.load_streams()
        RadioPanel._radiopanel = self

    @property
    def menu(self):
        """
        Gets a menu for the selected item
        :return: xlgui.widgets.menu.Menu or None if do not have it
        """
        model, it = self.tree.get_selection().get_selected()
        item = model[it][2]
        if isinstance(item, xl.playlist.Playlist):
            return self.playlist_menu
        elif isinstance(item, playlistpanel.TrackWrapper):
            return self.track_menu
        else:
            station = (
                item
                if isinstance(item, xl.radio.RadioStation)
                else (
                    item.station
                    if isinstance(item, (xl.radio.RadioList, xl.radio.RadioItem))
                    else None
                )
            )
            if station and hasattr(station, 'get_menu'):
                return station.get_menu(self)

    def load_streams(self):
        """
        Loads radio streams from plugins
        """
        for name in self.playlist_manager.playlists:
            pl = self.playlist_manager.get_playlist(name)
            if pl is not None:
                self.playlist_nodes[pl] = self.model.append(
                    self.custom, [self.playlist_image, pl.name, pl]
                )
                self._load_playlist_nodes(pl)
        self.tree.expand_row(self.model.get_path(self.custom), False)

        for name, value in self.manager.stations.items():
            self.add_driver(value)

    def _add_driver_cb(self, type, object, driver):
        self.add_driver(driver)

    def add_driver(self, driver):
        """
        Adds a driver to the radio panel
        """
        node = self.model.append(self.radio_root, [self.folder, str(driver), driver])
        self.nodes[driver] = node
        self.load_nodes[driver] = self.model.append(
            node, [self.refresh_image, _('Loading streams...'), None]
        )
        self.tree.expand_row(self.model.get_path(self.radio_root), False)

        if settings.get_option('gui/radio/%s_station_expanded' % driver.name, False):
            self.tree.expand_row(self.model.get_path(node), False)

    def _remove_driver_cb(self, type, object, driver):
        self.remove_driver(driver)

    def remove_driver(self, driver):
        """
        Removes a driver from the radio panel
        """
        if driver in self.nodes:
            self.model.remove(self.nodes[driver])
            del self.nodes[driver]

    def _setup_widgets(self):
        """
        Sets up the various widgets required for this panel
        """
        self.status = self.builder.get_object('status_label')

    @common.idle_add()
    def _set_status(self, message, timeout=0):
        self.status.set_text(message)

        if timeout:
            GLib.timeout_add_seconds(timeout, self._set_status, '', 0)

    def _connect_events(self):
        """
        Connects events used in this panel
        """

        self.builder.connect_signals(
            {'on_add_button_clicked': self._on_add_button_clicked}
        )
        self.tree.connect('row-expanded', self.on_row_expand)
        self.tree.connect('row-collapsed', self.on_collapsed)
        self.tree.connect('row-activated', self.on_row_activated)
        self.tree.connect('key-release-event', self.on_key_released)

        event.add_ui_callback(self._add_driver_cb, 'station_added', self.manager)
        event.add_ui_callback(self._remove_driver_cb, 'station_removed', self.manager)

    def _on_add_button_clicked(self, *e):
        dialog = dialogs.MultiTextEntryDialog(self.parent, _("Add Radio Station"))

        dialog.add_field(_("Name:"))
        url_field = dialog.add_field(_("URL:"))

        clipboard = Gtk.Clipboard.get(Gdk.SELECTION_CLIPBOARD)
        text = clipboard.wait_for_text()

        if text is not None:
            location = Gio.File.new_for_uri(text)

            if location.get_uri_scheme() is not None:
                url_field.set_text(text)

        result = dialog.run()
        dialog.hide()

        if result == Gtk.ResponseType.OK:
            name, uri = dialog.get_values()
            self._do_add_playlist(name, uri)

    @common.threaded
    def _do_add_playlist(self, name, uri):
        from xl import playlist, trax

        if playlist.is_valid_playlist(uri):
            pl = playlist.import_playlist(uri)
            pl.name = name
        else:
            pl = playlist.Playlist(name)
            tracks = trax.get_tracks_from_uri(uri)
            pl.extend(tracks)

        self.playlist_manager.save_playlist(pl)
        self._add_to_tree(pl)

    @common.idle_add()
    def _add_to_tree(self, pl):
        self.playlist_nodes[pl] = self.model.append(
            self.custom, [self.playlist_image, pl.name, pl]
        )
        self._load_playlist_nodes(pl)

    def _setup_tree(self):
        """
        Sets up the tree that displays the radio panel
        """
        box = self.builder.get_object('RadioPanel')
        self.tree = playlistpanel.PlaylistDragTreeView(self, True, True)
        self.tree.set_headers_visible(False)

        self.targets = [Gtk.TargetEntry.new('text/uri-list', 0, 0)]

        # columns
        text = Gtk.CellRendererText()
        if settings.get_option('gui/ellipsize_text_in_panels', False):
            from gi.repository import Pango

            text.set_property('ellipsize-set', True)
            text.set_property('ellipsize', Pango.EllipsizeMode.END)
        icon = Gtk.CellRendererPixbuf()
        col = Gtk.TreeViewColumn('radio')
        col.pack_start(icon, False)
        col.pack_start(text, True)
        col.set_attributes(icon, pixbuf=0)
        col.set_cell_data_func(text, self.cell_data_func)
        self.tree.append_column(col)

        self.model = Gtk.TreeStore(GdkPixbuf.Pixbuf, str, object)
        self.tree.set_model(self.model)

        self.track = icons.MANAGER.pixbuf_from_icon_name(
            'audio-x-generic', Gtk.IconSize.SMALL_TOOLBAR
        )
        self.folder = icons.MANAGER.pixbuf_from_icon_name(
            'folder', Gtk.IconSize.SMALL_TOOLBAR
        )
        self.refresh_image = icons.MANAGER.pixbuf_from_icon_name('view-refresh')

        self.custom = self.model.append(None, [self.folder, _("Saved Stations"), None])
        self.radio_root = self.model.append(
            None, [self.folder, _("Radio " "Streams"), None]
        )

        scroll = Gtk.ScrolledWindow()
        scroll.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
        scroll.add(self.tree)
        scroll.set_shadow_type(Gtk.ShadowType.IN)

        box.pack_start(scroll, True, True, 0)

    def on_row_activated(self, tree, path, column):
        item = self.model[path][2]
        if isinstance(item, xl.radio.RadioItem):
            self.emit('playlist-selected', item.get_playlist())
        elif isinstance(item, playlistpanel.TrackWrapper):
            self.emit('playlist-selected', item.playlist)
        elif isinstance(item, xl.playlist.Playlist):
            self.open_station(item)

    def open_station(self, playlist):
        """
        Opens a saved station
        """
        self.emit('playlist-selected', playlist)

    def get_menu(self):
        """
        Returns the menu that all radio stations use
        """
        m = menu.Menu(None)
        m.add_simple(_("Refresh"), self.on_reload, Gtk.STOCK_REFRESH)
        return m

    def on_key_released(self, widget, event):
        """
        Called when a key is released in the tree
        """
        if event.keyval == Gdk.KEY_Menu:
            mods, paths = self.tree.get_selection().get_selected_rows()
            if paths and paths[0]:
                iter = self.model.get_iter(paths[0])
                item = self.model.get_value(iter, 2)
                if isinstance(
                    item,
                    (xl.radio.RadioStation, xl.radio.RadioList, xl.radio.RadioItem),
                ):
                    if isinstance(item, xl.radio.RadioStation):
                        station = item
                    else:
                        station = item.station

                    if station and hasattr(station, 'get_menu'):
                        menu = station.get_menu(self)
                        menu.popup(event)
                elif isinstance(item, xl.playlist.Playlist):
                    Gtk.Menu.popup(
                        self.playlist_menu, None, None, None, None, 0, event.time
                    )
                elif isinstance(item, playlistpanel.TrackWrapper):
                    Gtk.Menu.popup(
                        self.track_menu, None, None, None, None, 0, event.time
                    )
            return True

        if event.keyval == Gdk.KEY_Left:
            mods, paths = self.tree.get_selection().get_selected_rows()
            if paths and paths[0]:
                self.tree.collapse_row(paths[0])
            return True

        if event.keyval == Gdk.KEY_Right:
            mods, paths = self.tree.get_selection().get_selected_rows()
            if paths and paths[0]:
                self.tree.expand_row(paths[0], False)
            return True

        return False

    def cell_data_func(self, column, cell, model, iter, user_data):
        """
        Called when the tree needs a value for column 1
        """
        object = model.get_value(iter, 1)
        cell.set_property('text', str(object))

    def drag_data_received(self, tv, context, x, y, selection, info, etime):
        """
        Called when someone drags some thing onto the playlist panel
        """
        # if the drag originated from radio view deny it
        # TODO this might change if we are allowed to change the order of radio
        if Gtk.drag_get_source_widget(context) == tv:
            context.drop_finish(False, etime)
            return

        locs = list(selection.get_uris())

        path = self.tree.get_path_at_pos(x, y)
        if path:
            # Add whatever we received to the playlist at path
            iter = self.model.get_iter(path[0])
            current_playlist = self.model.get_value(iter, 2)

            # if it's a track that we've dragged to, get the parent
            if isinstance(current_playlist, playlistpanel.TrackWrapper):
                current_playlist = current_playlist.playlist

            elif not isinstance(current_playlist, xl.playlist.Playlist):
                self._add_new_station(locs)
                return
            tracks, playlists = self.tree.get_drag_data(locs)
            current_playlist.extend(tracks)
            # Do we save in the case when a user drags a file onto a playlist in the playlist panel?
            # note that the playlist does not have to be open for this to happen
            self.playlist_manager.save_playlist(current_playlist, overwrite=True)
            self._load_playlist_nodes(current_playlist)
        else:
            self._add_new_station(locs)

    def _add_new_station(self, locs):
        """
        Add a new station
        """
        # If the user dragged files prompt for a new playlist name
        # else if they dragged a playlist add the playlist

        # We don't want the tracks in the playlists to be added to the
        # master tracks list so we pass in False
        tracks, playlists = self.tree.get_drag_data(locs, False)
        # First see if they dragged any playlist files
        for new_playlist in playlists:
            self.model.append(
                self.custom, [self.playlist_image, new_playlist.name, new_playlist]
            )
            # We are adding a completely new playlist with tracks so we save it
            self.playlist_manager.save_playlist(new_playlist, overwrite=True)

        # After processing playlist proceed to ask the user for the
        # name of the new playlist to add and add the tracks to it
        if len(tracks) > 0:
            dialog = dialogs.TextEntryDialog(
                _("Enter the name you want for your new playlist"), _("New Playlist")
            )
            result = dialog.run()
            if result == Gtk.ResponseType.OK:
                name = dialog.get_value()
                if not name == "":
                    # Create the playlist from all of the tracks
                    new_playlist = xl.playlist.Playlist(name)
                    new_playlist.extend(tracks)
                    self.playlist_nodes[new_playlist] = self.model.append(
                        self.custom,
                        [self.playlist_image, new_playlist.name, new_playlist],
                    )
                    self.tree.expand_row(self.model.get_path(self.custom), False)
                    # We are adding a completely new playlist with tracks so we save it
                    self.playlist_manager.save_playlist(new_playlist)
                    self._load_playlist_nodes(new_playlist)

    def drag_get_data(self, tv, context, selection_data, info, time):
        """
        Called when the user drags a playlist from the radio panel
        """
        tracks = self.tree.get_selected_tracks()

        if not tracks:
            return

        for track in tracks:
            DragTreeView.dragged_data[track.get_loc_for_io()] = track

        uris = trax.util.get_uris_from_tracks(tracks)
        selection_data.set_uris(uris)

    def drag_data_delete(self, *e):
        """
        stub
        """
        pass

    def on_reload(self, *e):
        """
        Called when the refresh button is clicked
        """
        selection = self.tree.get_selection()
        info = selection.get_selected_rows()
        if not info:
            return
        model, paths = info
        iter = self.model.get_iter(paths[0])
        object = self.model.get_value(iter, 2)

        try:
            self.loaded_nodes.remove(self.nodes[object])
        except ValueError:
            pass

        if isinstance(object, (xl.radio.RadioList, xl.radio.RadioStation)):
            self._clear_node(iter)
            self.load_nodes[object] = self.model.append(
                iter, [self.refresh_image, _("Loading streams..."), None]
            )

            self.complete_reload[object] = True
            self.tree.expand_row(self.model.get_path(iter), False)

    @staticmethod
    def set_station_expanded_value(station, value):
        settings.set_option('gui/radio/%s_station_expanded' % station, True)

    def on_row_expand(self, tree, iter, path):
        """
        Called when a user expands a row in the tree
        """
        driver = self.model.get_value(iter, 2)

        if not isinstance(driver, xl.playlist.Playlist):
            self.model.set_value(iter, 0, self.folder)

        if isinstance(driver, xl.radio.RadioStation) or isinstance(
            driver, xl.radio.RadioList
        ):
            if not self.nodes[driver] in self.loaded_nodes:
                self._load_station(iter, driver)

        if isinstance(driver, xl.radio.RadioStation):
            self.set_station_expanded_value(driver.name, True)

    def on_collapsed(self, tree, iter, path):
        """
        Called when someone collapses a tree item
        """
        driver = self.model.get_value(iter, 2)

        if not isinstance(driver, xl.playlist.Playlist):
            self.model.set_value(iter, 0, self.folder)

        if isinstance(driver, xl.radio.RadioStation):
            self.set_station_expanded_value(driver.name, False)

    @common.threaded
    def _load_station(self, iter, driver):
        """
        Loads a radio station
        """
        lists = None
        no_cache = False
        if driver in self.complete_reload:
            no_cache = True
            del self.complete_reload[driver]

        if isinstance(driver, xl.radio.RadioStation):
            try:
                lists = driver.get_lists(no_cache=no_cache)
            except RadioException as e:
                self._set_status(str(e), 2)
        else:
            try:
                lists = driver.get_items(no_cache=no_cache)
            except RadioException as e:
                self._set_status(str(e), 2)

        if not lists:
            return
        GLib.idle_add(self._done_loading, iter, driver, lists)

    def _done_loading(self, iter, object, items):
        """
        Called when an item is done loading.  Adds items to the tree
        """
        self.loaded_nodes.append(self.nodes[object])
        for item in items:
            if isinstance(item, xl.radio.RadioList):
                node = self.model.append(
                    self.nodes[object], [self.folder, item.name, item]
                )
                self.nodes[item] = node
                self.load_nodes[item] = self.model.append(
                    node, [self.refresh_image, _("Loading streams..."), None]
                )
            else:
                self.model.append(self.nodes[object], [self.track, item.name, item])

        try:
            self.model.remove(self.load_nodes[object])
            del self.load_nodes[object]
        except KeyError:
            pass

    def _clear_node(self, node):
        """
        Clears a node of all children
        """
        remove = []
        iter = self.model.iter_children(node)
        while iter:
            remove.append(iter)
            iter = self.model.iter_next(iter)
        for row in remove:
            self.model.remove(row)


def set_status(message, timeout=0):
    RadioPanel._radiopanel._set_status(message, timeout)



import os
from xlgui.preferences import widgets
from xl.nls import gettext as _

name = _("BPM Counter")
basedir = os.path.dirname(os.path.realpath(__file__))
ui = os.path.join(basedir, "bpm_prefs.ui")


class ShowConfimation(widgets.CheckPreference):
    default = True
    name = "plugin/bpm/show_confirmation_on_manual_setting"



'''
Routines useful for dealing with unicode data
'''

import locale
import logging
import string
import unicodedata

logger = logging.getLogger(__name__)


def shave_marks(text):
    """
    Removes diacritics from Latin characters and replaces them with their
    base characters

    :param text: Some input that will be converted to unicode string
    :returns: unicode string
    """
    text = str(text)
    decomposed_text = unicodedata.normalize('NFD', text)

    # Don't look for decomposed characters if there aren't any..
    if decomposed_text == text:
        return text

    keepers = []
    last = ' '
    for character in decomposed_text:
        if unicodedata.combining(character) and last in string.ascii_letters:
            continue  # Ignore diacritic on any Latin base character
        keepers.append(character)
        last = character
    shaved = ''.join(keepers)
    return unicodedata.normalize('NFC', shaved)


def to_unicode(x, encoding='utf-8', errors='strict'):
    """Force getting a unicode string from any object."""
    if isinstance(x, str):
        return x
    elif isinstance(x, bytes):
        return str(x, encoding, errors)
    else:
        return str(x)



# Copyright (C) 2008-2010 Adam Olsen
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.


from xl.metadata._base import BaseFormat
from mutagen import asf


class AsfFormat(BaseFormat):
    MutagenType = asf.ASF
    # TODO: figure out the the WM/ prefix is universal
    tag_mapping = {
        # fmt: off
        "artist"        : "Author",
        "album"         : "WM/AlbumTitle",
        "title"         : "Title",
        "genre"         : "WM/Genre",
        "tracknumber"   : "WM/TrackNumber",
        "date"          : "WM/Year",
        "albumartist"   : "WM/AlbumArtist",
        "grouping"      : "WM/ContentGroupDescription"
        # fmt: on
    }
    others = False
    writable = True

    def _get_tag(self, raw, tag_name):
        # the mutagen container for ASF returns the WM/ fields in its own
        # wrappers which are *almost* like a string.. convert them to
        # str so things don't break
        tag = super(AsfFormat, self)._get_tag(raw, tag_name)
        if isinstance(tag, list):
            attrs = [
                asf.ASFUnicodeAttribute,
                asf.ASFDWordAttribute,
                asf.ASFQWordAttribute,
                asf.ASFWordAttribute,
            ]

            def __process_tag(any_tag):
                for attrtype in attrs:
                    if isinstance(any_tag, attrtype):
                        return str(any_tag)
                return any_tag

            return [__process_tag(t) for t in tag]


# vim: et sts=4 sw=4



import http.client
import logging
from io import BytesIO

from spydaap.daap import DAAPError, DAAPObject, DAAPParseCodeTypes

log = logging.getLogger('daapclient')


class DAAPClient:
    def __init__(self):
        self.socket = None
        self.request_id = 0

    def connect(self, hostname, port=3689, password=None, user_agent=None):
        if self.socket is not None:
            raise DAAPError("DAAPClient: already connected.")
        #        if ':' in hostname:
        #            raise DAAPError('cannot connect to ipv6 addresses')
        # if it's an ipv6 address
        if ':' in hostname and hostname[0] != '[':
            hostname = '[' + hostname + ']'
        self.hostname = hostname
        self.port = port
        self.password = password
        self.user_agent = user_agent
        #        self.socket = httplib.HTTPConnection(hostname, port)
        self.socket = http.client.HTTPConnection(hostname + ':' + str(port))
        self.getContentCodes()  # practically required
        self.getInfo()  # to determine the remote server version

    def _get_response(self, r, params={}, gzip=1):
        """Makes a request, doing the right thing, returns the raw data"""

        if params:
            l = ['%s=%s' % (k, v) for k, v in params.items()]
            r = '%s?%s' % (r, '&'.join(l))

        log.debug('getting %s', r)

        headers = {'Client-DAAP-Version': '3.0', 'Client-DAAP-Access-Index': '2'}

        if self.user_agent:
            headers['User-Agent'] = self.user_agent

        if gzip:
            headers['Accept-encoding'] = 'gzip'

        if self.password:
            import base64

            b64 = base64.encodestring('%s:%s' % ('user', self.password))[:-1]
            headers['Authorization'] = 'Basic %s' % b64

        # TODO - we should allow for different versions of itunes - there
        # are a few different hashing algos we could be using. I need some
        # older versions of iTunes to test against.
        if self.request_id > 0:
            headers['Client-DAAP-Request-ID'] = self.request_id

        # headers[ 'Client-DAAP-Validation' ] = hash_v3(r, 2, self.request_id)

        # there are servers that don't allow >1 download from a single HTTP
        # session, or something. Reset the connection each time. Thanks to
        # Fernando Herrera for this one.
        self.socket.close()
        self.socket.connect()

        self.socket.request('GET', r, None, headers)

        response = self.socket.getresponse()
        return response

    def request(self, r, params={}, answers=1):
        """Make a request to the DAAP server, with the passed params. This
        deals with all the cikiness like validation hashes, etc, etc"""

        # this returns an HTTP response object
        response = self._get_response(r, params)
        status = response.status
        content = response.read()
        # if we got gzipped data base, gunzip it.
        if response.getheader("Content-Encoding") == "gzip":
            log.debug("gunzipping data")
            old_len = len(content)
            compressedstream = BytesIO(content)
            import gzip

            gunzipper = gzip.GzipFile(fileobj=compressedstream)
            content = gunzipper.read()
            log.debug("expanded from %s bytes to %s bytes", old_len, len(content))
        # close this, we're done with it
        response.close()

        if status == 401:
            raise DAAPError('DAAPClient: %s: auth required' % r)
        elif status == 403:
            raise DAAPError('DAAPClient: %s: Authentication failure' % r)
        elif status == 503:
            raise DAAPError(
                'DAAPClient: %s: 503 - probably max connections to server' % r
            )
        elif status == 204:
            # no content, ie logout messages
            return None
        elif status != 200:
            raise DAAPError(
                'DAAPClient: %s: Error %s making request' % (r, response.status)
            )

        return self.readResponse(content)

    def readResponse(self, data):
        """Convert binary response from a request to a DAAPObject"""
        data_str = BytesIO(data)
        daapobj = DAAPObject()
        daapobj.processData(data_str)
        return daapobj

    def getContentCodes(self):
        # make the request for the content codes
        response = self.request('/content-codes')
        # now parse and add this information to the dictionary
        DAAPParseCodeTypes(response)

    def getInfo(self):
        self.request('/server-info')

    def login(self):
        response = self.request("/login")
        sessionid = response.getAtom("mlid")
        if sessionid is None:
            log.debug('DAAPClient: login unable to determine session ID')
            return
        log.debug("Logged in as session %s", sessionid)
        return DAAPSession(self, sessionid)


class DAAPSession:
    def __init__(self, connection, sessionid):
        self.connection = connection
        self.sessionid = sessionid
        self.revision = 1

    def request(self, r, params={}, answers=1):
        """Pass the request through to the connection, adding the session-id
        parameter."""
        params['session-id'] = self.sessionid
        return self.connection.request(r, params, answers)

    def update(self):
        response = self.request("/update")
        self.revision = response.getAtom('musr')

    #        return response

    def databases(self):
        response = self.request("/databases")
        db_list = response.getAtom("mlcl").contains
        return [DAAPDatabase(self, d) for d in db_list]

    def library(self):
        # there's only ever one db, and it's always the library...
        return self.databases()[0]

    def logout(self):
        self.request("/logout")
        log.debug('DAAPSession: expired session id %s', self.sessionid)


# the atoms we want. Making this list smaller reduces memory footprint,
# and speeds up reading large libraries. It also reduces the metainformation
# available to the client.
daap_atoms = "dmap.itemid,dmap.itemname,daap.songalbum,daap.songartist,daap.songalbumartist,daap.songformat,daap.songtime,daap.songsize,daap.songgenre,daap.songyear,daap.songtracknumber,daap.songdiscnumber"


class DAAPDatabase:
    def __init__(self, session, atom):
        self.session = session
        self.name = atom.getAtom("minm")
        self.id = atom.getAtom("miid")

    def tracks(self):
        """returns all the tracks in this database, as DAAPTrack objects"""
        response = self.session.request(
            "/databases/%s/items" % self.id, {'meta': daap_atoms}
        )
        # response.printTree()
        track_list = response.getAtom("mlcl").contains
        return [DAAPTrack(self, t) for t in track_list]

    def playlists(self):
        response = self.session.request("/databases/%s/containers" % self.id)
        db_list = response.getAtom("mlcl").contains
        return [DAAPPlaylist(self, d) for d in db_list]


class DAAPPlaylist:
    def __init__(self, database, atom):
        self.database = database
        self.id = atom.getAtom("miid")
        self.name = atom.getAtom("minm")
        self.count = atom.getAtom("mimc")

    def tracks(self):
        """returns all the tracks in this playlist, as DAAPTrack objects"""
        response = self.database.session.request(
            "/databases/%s/containers/%s/items" % (self.database.id, self.id),
            {'meta': daap_atoms},
        )
        track_list = response.getAtom("mlcl").contains
        return [DAAPTrack(self.database, t) for t in track_list]


class DAAPTrack:
    attrmap = {
        'name': 'minm',
        'artist': 'asar',
        'album': 'asal',
        'id': 'miid',
        'type': 'asfm',
        'time': 'astm',
        'size': 'assz',
    }

    def __init__(self, database, atom):
        self.database = database
        self.atom = atom

    def __getattr__(self, name):
        if name in self.__dict__:
            return self.__dict__[name]
        elif name in DAAPTrack.attrmap:
            return self.atom.getAtom(DAAPTrack.attrmap[name])
        raise AttributeError(name)

    def request(self):
        """returns a 'response' object for the track's mp3 data.
        presumably you can stream from this or something"""

        # gotta bump this every track download
        self.database.session.connection.request_id += 1

        # get the raw response object directly, not the parsed version
        return self.database.session.connection._get_response(
            "/databases/%s/items/%s.%s" % (self.database.id, self.id, self.type),
            {'session-id': self.database.session.sessionid},
            gzip=0,
        )

    def save(self, filename):
        """saves the file to 'filename' on the local machine"""
        log.debug("saving to '%s'", filename)
        mp3 = open(filename, "wb")
        r = self.request()
        # doing this all on one lump seems to explode a lot. TODO - what
        # is a good block size here?
        data = r.read(32 * 1024)
        while data:
            mp3.write(data)
            data = r.read(32 * 1024)
        mp3.close()
        r.close()
        log.debug("Done")



from gi.repository import GLib
import threading
from xl import event

# nasty globals
on_ui_thread = [False]
calls = [0]


# TODO: monkeypatch instead
def glib_idle_add(fn, *args):
    was_ui_thread = on_ui_thread[0]
    on_ui_thread[0] = True

    try:
        fn(*args)
    finally:
        on_ui_thread[0] = was_ui_thread


GLib.idle_add = glib_idle_add


class NormalCallback:
    def __init__(self):
        self.called = False
        event.add_callback(self.on_cb, 'test')

    def destroy(self):
        event.remove_callback(self.on_cb, 'test')

    def on_cb(self, type, obj, data):
        self.called = True
        self.on_ui_thread = on_ui_thread[0]


class UiCallback:
    def __init__(self):
        self.called = False
        event.add_ui_callback(self.on_cb, 'test')

    def destroy(self):
        event.remove_callback(self.on_cb, 'test')

    def on_cb(self, type, obj, data):
        if on_ui_thread[0]:
            self.called = True


def _init_events():
    event.EVENT_MANAGER = event.EventManager()


def _finish_events():
    assert len(event.EVENT_MANAGER.callbacks) == 0
    assert len(event.EVENT_MANAGER.all_callbacks) == 0
    assert len(event.EVENT_MANAGER.ui_callbacks) == 0


def test_ui_events():
    _init_events()
    ucb = UiCallback()
    ncb = NormalCallback()

    on_ui_thread[0] = True
    event.log_event('test', ucb, None)

    assert ncb.called is True
    assert ncb.on_ui_thread is True
    assert ucb.called is True

    ucb.destroy()
    ncb.destroy()

    _finish_events()


def test_thread_events():
    _init_events()
    ucb = UiCallback()
    ncb = NormalCallback()

    def _run():
        on_ui_thread[0] = False
        event.log_event('test', ucb, None)

    t = threading.Thread(target=_run)
    t.start()
    t.join()

    assert ncb.called is True
    assert ncb.on_ui_thread is False
    assert ucb.called is True

    ucb.destroy()
    ncb.destroy()

    _finish_events()



import os, json
from gi.repository import Gtk
from gi.repository import GLib

from xl.nls import gettext as _
from xl.metadata import tags
from xlgui.widgets import common, dialogs
from xl import settings

name = _('Custom Collections')
basedir = os.path.dirname(os.path.realpath(__file__))
ui = os.path.join(basedir, "cco_pane.ui")


def init(dialog, builder):
    prefs = eco_prefs(dialog, builder)


class eco_prefs:
    def __init__(self, dialog, builder):
        self.builder = builder
        self.parent = dialog.parent

        self.grid = self.builder.get_object('preferences_pane')
        self.model = self.builder.get_object('model')
        self.list = self.builder.get_object('orders_tree')

        self.list.connect("row-activated", self._on_row_activated)
        self.builder.connect_signals(self)

        name_column = builder.get_object('name_column')

        """Name column"""
        name_cellrenderer = Gtk.CellRendererText()
        name_column.pack_start(name_cellrenderer, True)
        name_column.add_attribute(name_cellrenderer, 'text', 0)

        """Edit button"""
        edit_cellrenderer = common.ClickableCellRendererPixbuf()
        edit_cellrenderer.props.icon_name = 'edit'
        edit_cellrenderer.props.xalign = 1
        edit_cellrenderer.props.width = 30
        edit_cellrenderer.connect('clicked', self._on_edit_cellrenderer_clicked)
        name_column.pack_start(edit_cellrenderer, False)

        """Trash can"""
        remove_cellrenderer = common.ClickableCellRendererPixbuf()
        remove_cellrenderer.props.icon_name = 'edit-delete'
        remove_cellrenderer.props.xalign = 1
        remove_cellrenderer.props.width = 30
        remove_cellrenderer.connect('clicked', self._on_remove_cellrenderer_clicked)
        name_column.pack_start(remove_cellrenderer, False)

        self.custom_orders = custom_orders()

        self._grid_refresh()

    def _on_button_add_clicked(self, w):
        self._order_edit(-1)

    def _grid_refresh(self):
        self.list.set_model(None)
        self.model.clear()

        i = 0
        for order in self.custom_orders:
            it = self.model.append(None, (order['name'], i))
            i += 1

        self.list.set_model(self.model)

    def _on_row_activated(self, model, path, iter):
        order_number = self.model[path][1]
        self._order_edit(order_number)

    def _on_edit_cellrenderer_clicked(self, cellrenderer, path: str):
        order_number = self.model[path][1]
        GLib.idle_add(self._order_edit, order_number)

    def _on_remove_cellrenderer_clicked(self, cellrenderer, path: str):
        order_number = self.model[path][1]
        GLib.idle_add(self._ask_for_delete_order, order_number)

    def _ask_for_delete_order(self, order_number: int):
        check = dialogs.yesno(self.parent, _('Really delete this order?'))
        if check != Gtk.ResponseType.YES:
            return
        del self.custom_orders[order_number]
        self._grid_refresh()

    def _order_edit(self, order_number):
        if order_number > -1:
            order = self.custom_orders[order_number]
        else:
            order = {'name': "", 'levels': '', 'display': ''}

        dialog = dialogs.MultiTextEntryDialog(
            self.parent, _("Add or edit custom order")
        )

        tag_data = tags.tag_data
        usable_tags = ''
        for k, v in tag_data.items():
            if v:
                usable_tags = (
                    usable_tags + v.tag_name + ' (' + v.translated_name + ')\n'
                )

        tree_level_hint = _(
            'Comma separated list of tags to use as nodes in the tree view. Right now it\'s not possible to use more than one tag per node.\n'
            'Every tag can be used.'
        )
        display_hint = _(
            'Comma separated list of the tags to use for displaying single tracks as leaves in the tree. Tags are joined with a hyphen.\n'
            'Every tag can be used.'
        )

        dialog.add_field(_("Name:"), order['name'], _('Name of your order'))
        dialog.add_field(_("Tree Levels:"), order['levels'], tree_level_hint)
        dialog.add_field(_("Display:"), order['display'], display_hint)

        result = dialog.run()
        dialog.hide()

        if result != Gtk.ResponseType.OK:
            return

        name, levels, display = dialog.get_values()

        order['name'] = name
        order['levels'] = levels
        order['display'] = display

        if order_number > -1:
            self.custom_orders[order_number] = order
        else:
            self.custom_orders.append(order)

        self._grid_refresh()


class custom_orders(list):
    settings_name = 'cco/orders'

    def __init__(self):
        super().__init__(self)
        setting = settings.get_option(self.settings_name, None)
        if setting is not None:
            orders = json.loads(setting)
            for order in orders:
                super().append(order)

    def append(self, entry):
        super().append(entry)
        js = json.dumps(self)
        settings.set_option(self.settings_name, js)

    def __delitem__(self, arg):
        super().__delitem__(arg)
        js = json.dumps(self)
        settings.set_option(self.settings_name, js)

    def __setitem__(self, index, entry):
        super().__setitem__(index, entry)
        js = json.dumps(self)
        settings.set_option(self.settings_name, js)



from xlgui.preferences import widgets
from xl.nls import gettext as _
import os

name = _('DAAP Server')
basedir = os.path.dirname(os.path.realpath(__file__))
ui = os.path.join(basedir, 'daapserver_prefs.ui')


class PortPreference(widgets.SpinPreference):
    default = 3689
    name = 'plugin/daapserver/port'


class NamePreference(widgets.Preference):
    default = 'Exaile Share'
    name = 'plugin/daapserver/name'


class EnabledPreference(widgets.CheckPreference):
    default = True
    name = 'plugin/daapserver/enabled'


class HostPreference(widgets.Preference):
    default = '0.0.0.0'
    name = 'plugin/daapserver/host'



# moodbar - Replace Exaile's seekbar with a moodbar
# Copyright (C) 2015  Johannes Sasongko <sasongko@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import os
import subprocess
import threading
import tempfile

from gi.repository import Gio


class MoodbarGeneratorError(Exception):
    pass


class MoodbarGenerator:
    def check(self):
        """Check whether the generator works.

        For example, this can check for the presence of a `moodbar` executable.
        However, it does not guarantee that the moodbar generator will run
        without any errors during the generation process.

        :rtype: None
        :raise MoodbarGeneratorError: if the check fails
        """
        try:
            subprocess.check_call(('moodbar', '--help'), stdout=subprocess.DEVNULL)
        except subprocess.CalledProcessError:
            # executable exists but returned an error code
            pass
        except Exception:
            raise MoodbarGeneratorError(
                "Failed to run moodbar; make sure it is installed"
            )

    def generate(self, uri, callback=None):
        """
        :type uri: bytes
        :type callback: Callable[[bytes, bytes], None]
        :rtype: Optional[bytes]
        :raise MoodbarGeneratorError: on any error while generating moodbar
        """
        path = Gio.File.new_for_uri(uri).get_path()
        data = None
        if path:
            tmppath = f = None
            try:
                # Reserve a temporary file.
                fd, tmppath = tempfile.mkstemp(prefix='moodbar.')
                os.close(fd)

                cmd = ['moodbar', '-o', tmppath, path]
                subprocess.check_call(cmd)
                f = open(tmppath, 'rb')
                data = f.read()
            except Exception as e:
                # TODO propagate this error to UI, make sure to install required GStreamer plugins
                raise MoodbarGeneratorError(e)
            finally:
                if f:
                    f.close()
                if tmppath and os.path.exists(tmppath):
                    os.remove(tmppath)
        if callback:
            callback(uri, data)
        return data

    def generate_async(self, uri, callback=None):
        """
        :type uri: bytes
        :type callback: Callable[[bytes, bytes], None]
        """
        thread = threading.Thread(
            name=self.__class__.__name__, target=self.generate, args=(uri, callback)
        )
        thread.daemon = True
        thread.start()



# Copyright (C) 2008-2010 Adam Olsen
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.

"""
Devices

contains the DeviceManager and some generic Device classes
"""

from typing import Iterable
from xl import event, collection


class TransferNotSupportedError(Exception):
    pass


class Device:
    """
    a device

    must be subclassed for use
    """

    class_autoconnect = False
    library_class = collection.Library

    def __init__(self, name):
        self.name = name
        self.collection = collection.Collection(name=self.name)
        self.playlists = []
        self._connected = False
        self.transfer = None  # subclasses need to override this
        # if they want transferring

    def __get_connected(self):
        return self._connected

    def __set_connected(self, val):
        prior = self._connected
        self._connected = val
        if prior != val:
            if val:
                event.log_event("device_connected", self, self)
            else:
                event.log_event("device_disconnected", self, self)

    connected = property(__get_connected, __set_connected)

    def get_name(self):
        return self.name

    def autoconnect(self):
        if self.class_autoconnect:
            self.connect()

    def is_connected(self):
        return self.connected

    def connect(self):
        """
        connects to the device, creating Collections and Playlists
        as appropriate
        """
        raise NotImplementedError

    def disconnect(self):
        """
        disconnect from the device. should clear all stored metadata
        """
        raise NotImplementedError

    def get_collection(self):
        """
        returns the device's collection, if applicable
        """
        return self.collection

    def get_playlists(self):
        """
        returns a list of all playlists on the device, if any
        """
        return self.playlists

    def add_tracks(self, tracks):
        """
        Send tracks to the device
        """
        if not self.transfer:
            raise TransferNotSupportedError(
                "Device class does not " "support transfer."
            )
        self.transfer.enqueue(tracks)

    def start_transfer(self):
        if not self.transfer:
            raise TransferNotSupportedError(
                "Device class does not " "support transfer."
            )
        self.transfer.transfer()


class KeyedDevice(Device):
    """
    A utility class to inherit from that will return cached instances
    of your device if the device object is created with the same key.

    A device that inherits from this MUST have the key as the first
    argument to the __init__ function.

    @warning The __init__ function will be called again for devices
    that are created multiple times.
    """

    @staticmethod
    def __new__(cls, key):
        devices = getattr(cls, '__devices', {})

        device = devices.get(key, None)
        if device is None:
            device = Device.__new__(cls)
            device.__initialized = False
            device.__key = key
            devices[key] = device

        setattr(cls, '__devices', devices)
        return device

    def __init__(self, name):
        if self.__initialized:
            return

        # don't call this twice..
        Device.__init__(self, name)
        self.__initialized = True

    @classmethod
    def destroy(cls, device):
        """
        Call this to remove the device from the internal list
        """
        del getattr(cls, '__devices')[device.__key]


class DeviceManager:
    """
    manages devices
    """

    def __init__(self):
        self.devices = {}

    def add_device(self, device: Device):
        # make sure we don't overwrite existing devices
        count = 3
        if device.get_name() in self.devices:
            device.name += " (2)"
        while device.get_name() in self.devices:
            device.name = device.name[:-4] + " (%s)" % count
            count += 1

        self.devices[device.get_name()] = device
        event.log_event("device_added", self, device)

    def remove_device(self, device: Device):
        try:
            if device.connected:
                device.disconnect()
            del self.devices[device.get_name()]
        except KeyError:
            pass
        event.log_event("device_removed", self, device)

    def get_devices(self) -> Iterable[Device]:
        return self.devices.values()


# vim: et sts=4 sw=4



# Copyright (C) 2008-2010 Adam Olsen
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.


from gi.repository import GLib
from gi.repository import Gtk
from gi.repository import Pango

import logging
import time

from xl import common, event, player, settings, providers
from xl.common import classproperty
from xl.formatter import TrackFormatter
from xl.nls import gettext as _
from xlgui import icons
from xlgui.widgets import dialogs, rating, menu

logger = logging.getLogger(__name__)

DEFAULT_COLUMNS = ['tracknumber', 'title', 'album', 'artist', '__length']


class Column(Gtk.TreeViewColumn):
    name = ''
    display = ''
    menu_title = classproperty(lambda c: c.display)
    renderer = Gtk.CellRendererText
    formatter = classproperty(lambda c: TrackFormatter('$%s' % c.name))
    size = 10  # default size
    autoexpand = False  # whether to expand to fit space in Autosize mode
    datatype = str
    dataproperty = 'text'
    cellproperties = {}

    def __init__(self, container, player, font, size_ratio):
        if self.__class__ == Column:
            raise NotImplementedError(
                "Can't instantiate abstract class %r" % self.__class__
            )

        self._size_ratio = size_ratio
        self.container = container
        self.player = player
        self.settings_width_name = "gui/col_width_%s" % self.name
        self.cellrenderer = self.renderer()
        self.destroyed = False

        super().__init__(title=self.display)
        self.props.min_width = 3

        self.pack_start(self.cellrenderer, True)
        self.set_cell_data_func(self.cellrenderer, self.data_func)

        try:
            self.cellrenderer.set_property('font-desc', font)
        except TypeError:
            pass  # not all cells have a font

        try:
            self.cellrenderer.set_property('ellipsize', Pango.EllipsizeMode.END)
        except TypeError:  # cellrenderer doesn't do ellipsize - eg. rating
            pass

        for name, val in self.cellproperties.items():
            self.cellrenderer.set_property(name, val)

        self.set_reorderable(True)
        self.set_clickable(True)
        self.set_sizing(Gtk.TreeViewColumnSizing.FIXED)  # needed for fixed-height mode
        self.set_sort_order(Gtk.SortType.DESCENDING)

        # hack to allow button press events on the header to be detected
        self.set_widget(Gtk.Label(label=self.display))

        # Save the width of the column when it changes; save the notify id so
        # we don't emit an event when we're programmatically setting the width
        self._width_notify = self.connect('notify::width', self.on_width_changed)
        self._setup_sizing()

        event.add_ui_callback(
            self.on_option_set, "gui_option_set", destroy_with=container
        )

    def on_option_set(self, typ, obj, data):
        if data in ("gui/resizable_cols", self.settings_width_name):
            self._setup_sizing()

    def on_width_changed(self, column, wid):
        # Don't call get_width in the delayed function; it can be incorrect.
        # See https://github.com/exaile/exaile/issues/580
        self._delayed_width_changed(self.get_width())

    @common.glib_wait(100)
    def _delayed_width_changed(self, width):
        if not self.destroyed and width != settings.get_option(
            self.settings_width_name, -1
        ):
            settings.set_option(self.settings_width_name, width)

    def _setup_sizing(self):
        with self.handler_block(self._width_notify):
            if settings.get_option('gui/resizable_cols', False):
                self.set_resizable(True)
                self.set_expand(False)
                width = settings.get_option(self.settings_width_name, self.size)
                self.set_fixed_width(width)
            else:
                self.set_resizable(False)
                if self.autoexpand:
                    self.set_expand(True)
                    self.set_fixed_width(1)
                else:
                    self.set_expand(False)
                    self.set_fixed_width(self.size)

    def set_size_ratio(self, ratio):
        self._size_ratio = ratio

    def get_size_ratio(self):
        '''Returns how much bigger or smaller an icon should be'''
        return self._size_ratio

    def data_func(self, col, cell, model, iter, user_data):
        # warning: this function gets called from the render function, so do as
        #          little work as possible!

        cache = model.get_value(iter, 1)

        text = cache.get(self.name)
        if text is None:
            track = model.get_value(iter, 0)
            text = self.formatter.format(track)
            cache[self.name] = text

        cell.props.text = text

    def __repr__(self):
        return '%s(%r, %r, %r)' % (
            self.__class__.__name__,
            self.name,
            self.display,
            self.size,
        )


class EditableColumn(Column):
    renderer = Gtk.CellRendererText

    def __init__(self, *args):
        Column.__init__(self, *args)
        self.cellrenderer.connect('edited', self.on_edited)
        self.cellrenderer.connect('editing-started', self.on_editing_started)
        self.cellrenderer.connect('editing-canceled', self._reset_editable)

    def could_edit(self, treeview, path, event):
        # Returns True if an edit could take place

        # The 'aligned area' seems to correspond to where the text actually
        # is rendered. If the coordinates don't overlap that, then we don't
        # allow the edit to occur. Conveniently, this should deal with RTL
        # issues as well (not tested)
        cell_area = treeview.get_cell_area(path, self)

        # Ensure the renderer text attribute is updated before finding
        # it's width
        model = treeview.get_model()
        iter = model.get_iter(path)
        self.data_func(self, self.cellrenderer, model, iter, None)

        aa = self.cellrenderer.get_aligned_area(treeview, 0, cell_area)
        return aa.x <= event.x <= aa.x + aa.width

    def start_editing(self, treeview, path):
        # We normally keep the cell renderer's `editable` False so it doesn't
        # trigger editing on double-click. To actually start editing, we have
        # to temporarily set it to True until the user finishes editing.
        self.cellrenderer.props.editable = True
        treeview.set_cursor_on_cell(path, self, self.cellrenderer, start_editing=True)

    def _reset_editable(self, cellrenderer):
        cellrenderer.props.editable = False

    def on_edited(self, cellrenderer, path, new_text):
        self._reset_editable(cellrenderer)

        # Undo newline escaping
        new_text = new_text.replace('\\n', '\n')

        validate = getattr(self, 'validate', None)
        if validate:
            try:
                new_text = validate(new_text)
            except ValueError:
                return

        # Update the track
        model = self.get_tree_view().get_model()
        iter = model.get_iter(path)
        track = model.get_value(iter, 0)

        if not track.set_tag_raw(self.name, new_text):
            return

        # Invalidate/redraw the value immediately because we know
        # it's just a single change
        model.get_value(iter, 1).clear()
        model.row_changed(path, iter)

        if not track.write_tags():
            dialogs.error(
                None,
                "Error writing tags to %s"
                % GLib.markup_escape_text(track.get_loc_for_io()),
            )

    def on_editing_started(self, cellrenderer, editable, path):
        # Retrieve text in original form
        model = self.get_tree_view().get_model()

        iter = model.get_iter(path)
        track = model.get_value(iter, 0)

        text = getattr(self, 'edit_formatter', self.formatter).format(track)

        # Escape newlines
        text = text.replace('\n', '\\n')

        # Set text
        editable.set_text(text)

        if hasattr(self, 'validate'):
            editable.connect('changed', self.on_editing_changed)

    def on_editing_changed(self, w):
        try:
            self.validate(w.get_text())
        except ValueError:
            w.get_style_context().add_class('warning')
        else:
            w.get_style_context().remove_class('warning')


class TrackNumberColumn(Column):
    name = 'tracknumber'
    # TRANSLATORS: Title of the track number column
    display = _('#')
    menu_title = _('Track Number')
    size = 30
    cellproperties = {'xalign': 1.0, 'width-chars': 4}


providers.register('playlist-columns', TrackNumberColumn)


class TitleColumn(EditableColumn):
    name = 'title'
    display = _('Title')
    size = 200
    autoexpand = True


providers.register('playlist-columns', TitleColumn)


class ArtistColumn(EditableColumn):
    name = 'artist'
    display = _('Artist')
    size = 150
    autoexpand = True


providers.register('playlist-columns', ArtistColumn)


class AlbumArtistColumn(EditableColumn):
    name = 'albumartist'
    display = _('Album artist')
    size = 150
    autoexpand = True


providers.register('playlist-columns', AlbumArtistColumn)


class ComposerColumn(EditableColumn):
    name = 'composer'
    display = _('Composer')
    size = 150
    autoexpand = True


providers.register('playlist-columns', ComposerColumn)


class AlbumColumn(EditableColumn):
    name = 'album'
    display = _('Album')
    size = 150
    autoexpand = True


providers.register('playlist-columns', AlbumColumn)


class LengthColumn(Column):
    name = '__length'
    display = _('Length')
    size = 50
    cellproperties = {'xalign': 1.0}


providers.register('playlist-columns', LengthColumn)


class DiscNumberColumn(Column):
    name = 'discnumber'
    display = _('Disc')
    menu_title = _('Disc Number')
    size = 40
    cellproperties = {'xalign': 1.0, 'width-chars': 2}


providers.register('playlist-columns', DiscNumberColumn)


class RatingColumn(Column):
    name = '__rating'
    display = _('Rating')
    renderer = rating.RatingCellRenderer
    dataproperty = 'rating'
    cellproperties = {'follow-state': False}

    def __init__(self, *args):
        Column.__init__(self, *args)
        self.cellrenderer.connect('rating-changed', self.on_rating_changed)
        self.cellrenderer.size_ratio = self.get_size_ratio()
        self.saved_model = None

    def data_func(self, col, cell, model, iter, user_data):
        track = model.get_value(iter, 0)
        cell.props.rating = track.get_rating()
        self.saved_model = model

    def __get_size(self):
        """
        Retrieves the optimal size
        """
        size = icons.MANAGER.pixbuf_from_rating(0, self.get_size_ratio()).get_width()
        size += 2  # FIXME: Find the source of this

        return size

    size = property(__get_size)

    def on_rating_changed(self, widget, path, rating):
        """
        Updates the rating of the selected track
        """
        iter = self.saved_model.get_iter(path)
        track = self.saved_model.get_value(iter, 0)
        oldrating = track.get_rating()

        if rating == oldrating:
            rating = 0

        rating = track.set_rating(rating)
        event.log_event('rating_changed', self, rating)


providers.register('playlist-columns', RatingColumn)


class DateColumn(Column):
    name = 'date'
    display = _('Date')
    size = 50


providers.register('playlist-columns', DateColumn)


class YearColumn(Column):
    name = 'year'
    display = _('Year')
    size = 45


providers.register('playlist-columns', YearColumn)


class GenreColumn(EditableColumn):
    name = 'genre'
    display = _('Genre')
    size = 100
    autoexpand = True


providers.register('playlist-columns', GenreColumn)


class BitrateColumn(Column):
    name = '__bitrate'
    display = _('Bitrate')
    size = 45
    cellproperties = {'xalign': 1.0}


providers.register('playlist-columns', BitrateColumn)


class IoLocColumn(Column):
    name = '__loc'
    display = _('Location')
    size = 200
    autoexpand = True


providers.register('playlist-columns', IoLocColumn)


class FilenameColumn(Column):
    name = '__basename'
    display = _('Filename')
    size = 200
    autoexpand = True


providers.register('playlist-columns', FilenameColumn)


class PlayCountColumn(Column):
    name = '__playcount'
    display = _('Playcount')
    size = 50
    cellproperties = {'xalign': 1.0}


providers.register('playlist-columns', PlayCountColumn)


class BPMColumn(EditableColumn):
    name = 'bpm'
    display = _('BPM')
    size = 40
    cellproperties = {'xalign': 1.0, 'editable': True}
    validate = lambda s, v: int(v)


providers.register('playlist-columns', BPMColumn)


class LanguageColumn(EditableColumn):
    name = 'language'
    display = _('Language')
    size = 100
    autoexpand = True


providers.register('playlist-columns', LanguageColumn)


class LastPlayedColumn(Column):
    name = '__last_played'
    display = _('Last played')
    size = 80


providers.register('playlist-columns', LastPlayedColumn)


class DateAddedColumn(Column):
    name = '__date_added'
    display = _('Date added')
    size = 80


providers.register('playlist-columns', DateAddedColumn)


class ScheduleTimeColumn(Column):
    name = 'schedule_time'
    display = _('Schedule')
    size = 80

    def __init__(self, *args):
        Column.__init__(self, *args)
        self.timeout_id = None

        event.add_ui_callback(
            self.on_queue_current_playlist_changed,
            'queue_current_playlist_changed',
            player.QUEUE,
        )
        event.add_ui_callback(
            self.on_playback_player_start, 'playback_player_start', player.PLAYER
        )
        event.add_ui_callback(
            self.on_playback_player_end, 'playback_player_end', player.PLAYER
        )

    def data_func(self, col, cell, model, iter, user_data):
        """
        Sets the schedule time if appropriate
        """
        text = None
        playlist = self.container.playlist

        # Only display the schedule time if
        # 1) there is playback at all
        # 2) this playlist is currently played
        # 3) playback is not in shuffle mode
        # 4) the current track is not repeated
        if (
            not self.player.is_stopped()
            and playlist is self.player.queue.current_playlist
            and playlist.shuffle_mode == 'disabled'
            and playlist.repeat_mode != 'track'
        ):
            track = model.get_value(iter, 0)
            position = playlist.index(track)
            current_position = playlist.current_position

            # 5) this track is after the currently played one
            if position > current_position:
                # The delay is the accumulated length of all tracks
                # between the currently playing and this one
                try:
                    delay = sum(
                        t.get_tag_raw('__length')
                        for t in playlist[current_position:position]
                    )
                except TypeError:
                    # on tracks with length == None, we cannot determine
                    # when later tracks will play
                    pass
                else:
                    # Subtract the time which already has passed
                    delay -= self.player.get_time()
                    # The schedule time is the current time plus delay
                    schedule_time = time.localtime(time.time() + delay)
                    text = time.strftime('%H:%M', schedule_time)

        cell.props.text = text

    def start_timer(self):
        """
        Enables realtime updates
        """
        timeout_id = self.timeout_id

        # Make sure to stop any timer still running
        if timeout_id is not None:
            GLib.source_remove(timeout_id)

        self.timeout_id = GLib.timeout_add_seconds(60, self.on_timeout)

    def stop_timer(self):
        """
        Disables realtime updates
        """
        timeout_id = self.timeout_id

        if timeout_id is not None:
            GLib.source_remove(timeout_id)
            self.timeout_id = None

        # Update once more
        self.on_timeout()

    def on_timeout(self):
        """
        Makes sure schedule times are updated in realtime
        """
        self.queue_resize()
        view = self.get_tree_view()
        if view is not None:
            view.queue_draw()
            return True

        self.timeout_id = None
        return False

    def on_queue_current_playlist_changed(self, e, queue, current_playlist):
        """
        Disables realtime updates for all playlists
        and re-enables them for the current playlist
        """
        self.stop_timer()

        if current_playlist is self.container.playlist:
            self.start_timer()

    def on_playback_player_start(self, e, player, track):
        """
        Enables realtime updates for the current playlist
        """
        self.stop_timer()

        if self.player.queue.current_playlist is self.container.playlist:
            logger.debug('Playback started, enabling realtime updates')
            self.start_timer()

    def on_playback_player_end(self, e, player, track):
        """
        Disables realtime updates for all playlists
        """
        self.stop_timer()


providers.register('playlist-columns', ScheduleTimeColumn)


class CommentColumn(EditableColumn):
    name = 'comment'
    display = _('Comment')
    size = 200
    autoexpand = True
    # Remove the newlines to fit into the vertical space of rows
    formatter = TrackFormatter('${comment:newlines=strip}')
    edit_formatter = TrackFormatter('$comment')
    cellproperties = {'editable': True}


providers.register('playlist-columns', CommentColumn)


class GroupingColumn(EditableColumn):
    name = 'grouping'
    display = _('Grouping')
    size = 200
    autoexpand = True


providers.register('playlist-columns', GroupingColumn)


def _validate_time(self, v):
    rv = 0
    m = 1
    for i in reversed(v.split(':')):
        i = int(i)
        if i < 0 or i > 59:
            raise ValueError
        rv += i * m
        m *= 60
    return rv


class StartOffsetColumn(EditableColumn):
    name = '__startoffset'
    display = _('Start Offset')
    size = 50
    cellproperties = {'xalign': 1.0, 'editable': True}
    validate = _validate_time


providers.register('playlist-columns', StartOffsetColumn)


class StopOffsetColumn(EditableColumn):
    name = '__stopoffset'
    display = _('Stop Offset')
    size = 50
    cellproperties = {'xalign': 1.0, 'editable': True}
    validate = _validate_time


providers.register('playlist-columns', StopOffsetColumn)


class WebsiteColumn(EditableColumn):
    name = 'website'
    display = _('Website')
    size = 200
    autoexpand = True


providers.register('playlist-columns', WebsiteColumn)


class ColumnMenuItem(menu.MenuItem):
    """
    A menu item dedicated to display the
    status of a column and change it
    """

    def __init__(self, column, after=None):
        """
        Sets up the menu item from a column description

        :param column: the playlist column
        :type column: :class:`Column`
        :param after: enumeration of menu
            items before this one
        :type after: list of strings
        """
        menu.MenuItem.__init__(self, column.name, self.factory, after)
        self.title = column.menu_title

    def factory(self, menu, parent, context):
        """
        Creates the menu item
        """
        item = Gtk.CheckMenuItem.new_with_label(self.title)
        active = self.is_selected(self.name, parent, context)
        item.set_active(active)
        item.connect('activate', self.on_item_activate, self.name, parent, context)

        return item

    def is_selected(self, name, parent, context):
        """
        Returns whether a column is selected

        :rtype: bool
        """
        return name in settings.get_option('gui/columns', DEFAULT_COLUMNS)

    def on_item_activate(self, menu_item, name, parent, context):
        """
        Updates the columns setting
        """
        columns = settings.get_option('gui/columns', DEFAULT_COLUMNS)

        if name in columns:
            columns.remove(name)
        elif 'trigger_column' in context and context['trigger_column']:
            col_name = context['trigger_column'].name
            s = columns.index(col_name) + 1
            columns.insert(s, name)
        else:
            columns.append(name)

        settings.set_option('gui/columns', columns)


def __register_playlist_columns_menuitems():
    """
    Registers standard menu items for playlist columns
    """

    def is_column_selected(name, parent, context):
        """
        Returns whether a menu item should be checked
        """
        return name in settings.get_option('gui/columns', DEFAULT_COLUMNS)

    def is_resizable(name, parent, context):
        """
        Returns whether manual or automatic sizing is requested
        """
        resizable = settings.get_option('gui/resizable_cols', False)

        if name == 'resizable':
            return resizable
        elif name == 'autosize':
            return not resizable

    def on_column_item_activate(menu_item, name, parent, context):
        """
        Updates columns setting
        """
        columns = settings.get_option('gui/columns', DEFAULT_COLUMNS)

        if name in columns:
            columns.remove(name)
        else:
            columns.append(name)

        settings.set_option('gui/columns', columns)

    def on_sizing_item_activate(menu_item, name, parent, context):
        """
        Updates column sizing setting
        """

        # Ignore the activation if the menu item is not actually active
        if not menu_item.get_active():
            return

        if name == 'resizable':
            settings.set_option('gui/resizable_cols', True)
        elif name == 'autosize':
            settings.set_option('gui/resizable_cols', False)

    columns = [
        'tracknumber',
        'title',
        'artist',
        'album',
        '__length',
        'genre',
        '__rating',
        'date',
    ]

    for provider in providers.get('playlist-columns'):
        if provider.name not in columns:
            columns += [provider.name]

    menu_columns = []
    for name in columns:
        column = providers.get_provider('playlist-columns', name)
        menu_columns += [column]

    # Sort menu alphabetical
    menu_columns.sort(key=lambda elem: elem.menu_title)

    menu_items = []
    after = []
    for column in menu_columns:
        menu_item = ColumnMenuItem(column, after)
        menu_items += [menu_item]
        after = [menu_item.name]

    separator_item = menu.simple_separator('columns_separator', after)
    menu_items += [separator_item]
    after = [separator_item.name]

    sizing_item = menu.radio_menu_item(
        'resizable',
        after,
        _('_Resizable'),
        'column-sizing',
        is_resizable,
        on_sizing_item_activate,
    )
    menu_items += [sizing_item]
    after = [sizing_item.name]

    sizing_item = menu.radio_menu_item(
        'autosize',
        after,
        _('_Autosize'),
        'column-sizing',
        is_resizable,
        on_sizing_item_activate,
    )
    menu_items += [sizing_item]

    for menu_item in menu_items:
        providers.register('playlist-columns-menu', menu_item)


__register_playlist_columns_menuitems()



# Copyright (C) 2008-2010 Adam Olsen
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.

from gi.repository import Gio
from gi.repository import Gtk
import logging

import xl.collection
from xl.nls import gettext as _
from xl import xdg
from xlgui.widgets import dialogs
from xlgui.guiutil import GtkTemplate

logger = logging.getLogger(__name__)


@GtkTemplate('ui', 'collection_manager.ui')
class CollectionManagerDialog(Gtk.Dialog):
    """
    Allows you to choose which directories are in your library
    """

    __gtype_name__ = 'CollectionManager'

    view, model, remove_button, content_area = GtkTemplate.Child.widgets(4)
    location_column, location_cellrenderer = GtkTemplate.Child.widgets(2)

    def __init__(self, parent, collection: xl.collection.Collection):
        """
        Initializes the dialog
        """
        Gtk.Dialog.__init__(self)
        self.init_template()

        self.parent = parent
        self.collection = collection

        self.set_transient_for(self.parent)
        self.message = dialogs.MessageBar(
            parent=self.content_area, buttons=Gtk.ButtonsType.CLOSE
        )

        for location, library in collection.libraries.items():
            self.model.append([location, library.monitored, library.startup_scan])

        # Override the data function for location_column, so that it
        # displays parsed location names instead of escaped URIs
        def display_parsed_location(column, cell, model, iter, *data):
            location_uri = model.get(iter, 0)[0]  # first column
            location = Gio.File.new_for_uri(location_uri)
            cell.set_property("text", location.get_parse_name())

        self.location_column.set_cell_data_func(
            self.location_cellrenderer, display_parsed_location
        )

    def get_items(self):
        """
        Returns the items in the dialog
        """
        items = []

        for row in self.model:
            items += [(row[0], row[1], row[2])]

        return items

    @GtkTemplate.Callback
    def on_monitored_cellrenderer_toggled(self, cell, path):
        """
        Enables or disables monitoring
        """
        monitored = not cell.get_active()
        cell.set_active(monitored)
        self.model[path][1] = monitored

    @GtkTemplate.Callback
    def on_startup_cellrenderer_toggled(self, cell, path):
        """
        Enables or disables scanning on startup
        """
        if self.model[path][1]:
            scan_on_startup = not cell.get_active()
            cell.set_active(scan_on_startup)
            self.model[path][2] = scan_on_startup

    @GtkTemplate.Callback
    def on_add_button_clicked(self, widget):
        """
        Adds a path to the list
        """
        dialog = Gtk.FileChooserDialog(
            _("Add a Directory"),
            self.parent,
            Gtk.FileChooserAction.SELECT_FOLDER,
            (
                Gtk.STOCK_CANCEL,
                Gtk.ResponseType.CANCEL,
                Gtk.STOCK_ADD,
                Gtk.ResponseType.OK,
            ),
        )
        dialog.set_current_folder(xdg.get_last_dir())
        dialog.set_local_only(False)  # enable gio
        response = dialog.run()

        # XXX: For some reason, on Ubuntu 12.10 (GTK 2.24.13), hiding the
        # dialog before retrieving the results causes an incorrect URI
        # to be retrieved.

        uri = dialog.get_uri()
        dialog.hide()

        if response == Gtk.ResponseType.OK:
            location = Gio.File.new_for_uri(uri)
            removals = []

            for row in self.model:
                library_location = Gio.File.new_for_uri(row[0])
                # monitored = row[1]
                # scan_on_startup = row[2]

                if location.equal(library_location) or location.has_prefix(
                    library_location
                ):
                    self.message.show_warning(
                        _('Directory not added.'),
                        _(
                            'The directory is already in your collection '
                            'or is a subdirectory of another directory in '
                            'your collection.'
                        ),
                    )
                    break
                elif library_location.has_prefix(location):
                    removals += [row.iter]
            else:
                self.model.append([location.get_uri(), False, False])

                for iter in removals:
                    self.model.remove(iter)

        dialog.destroy()

    @GtkTemplate.Callback
    def on_remove_button_clicked(self, widget):
        """
        removes a path from the list
        """
        selection = self.view.get_selection()
        model, iter = selection.get_selected()
        model.remove(iter)

    @GtkTemplate.Callback
    def on_rescan_button_clicked(self, widget):
        """
        Triggers rescanning the collection
        """

        from xlgui import main

        main.mainwindow().controller.on_rescan_collection()

    @GtkTemplate.Callback
    def on_force_rescan_button_clicked(self, widget):
        """
        Triggers a slow rescan of the collection
        """

        from xlgui import main

        main.mainwindow().controller.on_rescan_collection_forced()

    @GtkTemplate.Callback
    def on_selection_changed(self, selection):
        """
        Enables or disables the "Remove" button
        """
        rows_selected = selection.count_selected_rows() > 0
        self.remove_button.set_sensitive(rows_selected)



import os
from xl.nls import gettext as _
from xlgui.preferences import widgets

name = _("Pause on Screensaver")
basedir = os.path.dirname(os.path.realpath(__file__))
ui = os.path.join(basedir, 'prefs.ui')


class Unpause(widgets.CheckPreference):
    default = False
    name = 'screensaverpause/unpause'



# screensaverpause - pauses Exaile playback on screensaver activation
# Copyright (C) 2009-2011  Johannes Sasongko <sasongko@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.

from gi.repository import GLib

import dbus
from xl import event, player, settings

SERVICES = [
    dict(  # GNOME
        bus_name='org.gnome.ScreenSaver',
        path='/org/gnome/ScreenSaver',
        dbus_interface='org.gnome.ScreenSaver',
    ),
    dict(  # MATE
        bus_name='org.mate.ScreenSaver',
        path='/org/mate/ScreenSaver',
        dbus_interface='org.mate.ScreenSaver',
    ),
    dict(  # CINNAMON
        bus_name='org.cinnamon.ScreenSaver',
        path='/org/cinnamon/ScreenSaver',
        dbus_interface='org.cinnamon.ScreenSaver',
    ),
    dict(  # XFCE
        bus_name='org.xfce.ScreenSaver',
        path='/org/xfce/ScreenSaver',
        dbus_interface='org.xfce.ScreenSaver',
    ),
    dict(  # KDE
        bus_name='org.freedesktop.ScreenSaver',
        path='/',
        dbus_interface='org.freedesktop.ScreenSaver',
    ),
]

from . import prefs


def get_preferences_pane():
    return prefs


matches = set()
bus = None
was_playing = None


def screensaver_active_changed(is_active):
    global was_playing
    if is_active:
        was_playing = player.PLAYER.is_playing()
        player.PLAYER.pause()
    elif was_playing and settings.get_option("screensaverpause/unpause", 0):
        player.PLAYER.unpause()


def enable(exaile):
    if exaile.loading:
        event.add_callback(_enable, 'exaile_loaded')
    else:
        _enable()


def _enable(*a):
    global bus
    bus = dbus.SessionBus()
    for service in SERVICES:
        matches.add(
            bus.add_signal_receiver(
                screensaver_active_changed, signal_name='ActiveChanged', **service
            )
        )


def disable(exaile):
    if bus is None:
        return
    for match in frozenset(matches):
        match.remove()
        matches.remove(match)


def test():
    import dbus.mainloop.glib as dbgl

    dbgl.DBusGMainLoop(set_as_default=True)

    global bus
    bus = dbus.SessionBus()

    for service in SERVICES:
        try:
            proxy = bus.get_object(
                service['bus_name'], service['path'], follow_name_owner_changes=True
            )
        except dbus.DBusException:
            continue
        break
    else:
        return None
    assert proxy
    interface = dbus.Interface(proxy, service['dbus_interface'])
    mainloop = GLib.MainLoop()

    def active_changed(new_value):
        if not new_value:
            mainloop.quit()

    interface.connect_to_signal('ActiveChanged', screensaver_active_changed)

    # For some reason Lock never returns.
    interface.Lock(ignore_reply=True)

    mainloop.run()


if __name__ == '__main__':
    test()


# vi: et sts=4 sw=4 tw=80



# Copyright (C) 2009-2010 Aren Olson
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.

import logging
import threading
import os

from gi.repository import GLib

from xl import common, event, formatter, settings, transcoder, trax
from xl.nls import gettext as _
from xlgui.panel import device
import xlgui

logger = logging.getLogger(__name__)


# TODO: Improve progress update
# TODO: Ask for folder where to put the disc instead of blindly copying it
# TODO: add eject button
# TODO: allow to directly play from panel


class CDImportThread(common.ProgressThread):
    def __init__(self, cd_importer):
        common.ProgressThread.__init__(self)

        self.cd_importer = cd_importer

    def stop(self):
        """
        Stops the thread
        """
        self.cd_importer.stop()
        common.ProgressThread.stop(self)

    def on_progress_update(self, progress=None):
        """
        Notifies about progress changes
        """
        if progress is None:
            progress = self.cd_importer.get_progress() * 100

        self.emit('progress-update', int(progress))

        return True

    def run(self):
        """
        Runs the thread
        """
        progress_id = GLib.timeout_add_seconds(1, self.on_progress_update)
        self.cd_importer.do_import()
        GLib.source_remove(progress_id)
        self.emit('done')


class CDPanel(device.FlatPlaylistDevicePanel):
    def __init__(self, *args):
        device.FlatPlaylistDevicePanel.__init__(self, *args)
        self.__importing = False

        event.add_ui_callback(self._tree_queue_draw, 'cd_info_retrieved')

    def _tree_queue_draw(self, _event_type, cdplaylist, disc_title=None):
        if not hasattr(self.fppanel, 'tree'):
            return

        # TODO: set name to panel. This requires a GUI change in
        # panel/flatplaylist/FlatPlaylistPanel respectively flatplaylist.ui
        # Do not use self.name because it breaks the panel after switching to another CD
        if disc_title is None:
            self.panel_title = _('Unknown disc')
        else:
            self.panel_title = disc_title

        if cdplaylist in self.device.playlists:
            logger.info("Calling queue_draw for %s", str(cdplaylist))
            self.fppanel.tree.queue_draw()

    def do_import(self, tracks):
        if self.__importing:
            return
        self.__importing = True
        cd_importer = CDImporter(tracks)
        thread = CDImportThread(cd_importer)
        thread.connect('done', lambda *e: self._import_finish())
        self.main.controller.progress_manager.add_monitor(
            thread, _("Importing CD..."), 'drive-optical'
        )

    def _import_finish(self):
        self.__importing = False


class CDImporter:
    def __init__(self, tracks):
        self.tracks = [t for t in tracks if t.get_loc_for_io().startswith("cdda")]
        self.duration = sum(t.get_tag_raw('__length') for t in self.tracks)
        self.formatter = formatter.TrackFormatter(
            settings.get_option(
                "cd_import/outpath",
                "%s/$artist/$album/$tracknumber - $title" % os.getenv("HOME"),
            )
        )
        self.current = None
        self.current_len = None
        self.progress = 0.0

        self.running = False

        self.cont = None
        self.__prepare_transcoder()

    def __prepare_transcoder(self):
        formats = transcoder.get_formats()
        default_format = next(iter(formats))
        self.format = settings.get_option("cd_import/format", default_format)
        default_quality = formats[default_format]['default']
        self.quality = settings.get_option("cd_import/quality", default_quality)
        self.transcoder = transcoder.Transcoder(
            self.format, self.quality, self._error_cb, self._end_cb
        )

    def do_import(self):
        self.running = True

        self.cont = threading.Event()

        self.transcoder.set_format(self.format)
        if self.quality != -1:
            self.transcoder.set_quality(self.quality)

        for tr in self.tracks:
            self.cont.clear()
            self.current = tr
            self.current_len = tr.get_tag_raw('__length')
            loc = tr.get_loc_for_io()
            trackno, device = loc[7:].split("/#")
            src = "cdparanoiasrc track=%s device=\"%s\"" % (trackno, device)
            self.transcoder.set_raw_input(src)
            outloc = self.get_output_location(tr)
            self.transcoder.set_output(outloc)
            self.transcoder.start_transcode()
            self.cont.wait()
            if not self.running:
                break
            tr2 = trax.Track("file://" + outloc)
            ntags = {
                t: tr.get_tag_raw(t) for t in tr.list_tags() if not t.startswith("__")
            }
            tr2.set_tags(**ntags)
            tr2.write_tags()
            try:
                incr = tr.get_tag_raw('__length') / self.duration
                self.progress += incr
            except Exception:
                raise
        self.progress = 100.0

    def _end_cb(self):
        self.cont.set()
        xlgui.main.mainwindow().message.show_info(_("Finished transcoding files"))

    def _error_cb(self, gerror, message_string):
        self.running = False
        xlgui.main.mainwindow().message.show_error(
            _("Error transcoding files from CD."), "%s" % gerror.message.encode()
        )

    def get_output_location(self, track):
        path = self.formatter.format(track)
        directorypath = os.path.dirname(path)

        if not os.path.exists(directorypath):
            os.makedirs(directorypath)

        extension = transcoder.FORMATS[self.transcoder.dest_format]['extension']

        return path + '.' + extension

    def stop(self):
        self.running = False
        self.transcoder.stop()

    def get_progress(self):
        if not self.current or not self.current_len:
            return self.progress
        incr = self.current_len / self.duration
        pos = self.transcoder.get_time() / self.current_len
        return self.progress + pos * incr



import os
import sys

# - good = works
#   - main = useful
#   - extra = not very useful
# - bad = doesn't work

EXTRA = ['helloworld']
BAD = []


def scan():
    all = set(f for f in os.listdir('.') if os.path.isdir(f))
    bad = set(BAD)
    good = all - bad
    extra = set(EXTRA)
    main = good - extra
    return locals()


plugins = scan()


def parse(argv):
    if len(argv) == 1:
        return plugins['all']
    return plugins[argv[1]]


if __name__ == '__main__':
    names = sorted(parse(sys.argv))
    print((' '.join(names)))



from xl.nls import gettext


def N_(x):
    return x


class _TD:
    __slots__ = [
        'name',  # descriptive name
        'translated_name',  # translated name
        'tag_name',  # raw tag name
        'type',
        'editable',
        'min',
        'max',
        'use_disk',  # set true if should retrieve tag from disk -- which means
        # the tag cannot be stored in the database
    ]

    def __init__(self, name, type, **kwargs):
        self.name = name
        self.translated_name = gettext(name)
        self.type = type

        # these are overridable by keyword arg
        self.editable = True
        self.use_disk = False

        for k, v in kwargs.items():
            setattr(self, k, v)


#: List of metadata tags currently supported by exaile, which are
#: translated into the corresponding tag for each audio format if
#: it supports it. Prefer to extend this list's features instead
#: of creating your own list of tag metadata
#:
#: @note We use N_ (fake gettext) because for some uses these strings are
#: translated later, so we store both the translated and untranslated
#: version

tag_data = {
    # fmt: off
    'album':            _TD(N_('Album'),        'text'),
    'arranger':         _TD(N_('Arranger'),     'text'),
    'artist':           _TD(N_('Artist'),       'text'),
    'albumartist':      _TD(N_('Album artist'), 'text'),
    'author':           _TD(N_('Author'),       'text'),
    'bpm':              _TD(N_('BPM'),          'int', min=0, max=500),
    'copyright':        _TD(N_('Copyright'),    'text'),
    'comment':          _TD(N_('Comment'),      'multiline'),
    'composer':         _TD(N_('Composer'),     'text'),
    'conductor':        _TD(N_('Conductor'),    'text'),
    'cover':            _TD(N_('Cover'),        'image', use_disk=True),
    'date':             _TD(N_('Date'),         'datetime'),
    'discnumber':       _TD(N_('Disc'),         'dblnum', min=0, max=50),
    'encodedby':        _TD(N_('Encoded by'),   'text'),
    'genre':            _TD(N_('Genre'),        'text'),
    'grouping':         _TD(N_('Grouping'),     'text'),
    'isrc':             _TD(N_('ISRC'),         'text'),
    'language':         _TD(N_('Language'),     'text'),
    'lyrics':           _TD(N_('Lyrics'),       'multiline', use_disk=True),
    'lyricist':         _TD(N_('Lyricist'),     'text'),
    'organization':     _TD(N_('Organization'), 'text'),
    'originalalbum':    _TD(N_('Original album'), 'text'),
    'originalartist':   _TD(N_('Original artist'), 'text'),
    'originaldate':     _TD(N_('Original date'), 'text'),
    'part':             None,
    'performer':        _TD(N_('Performer'),    'text'),
    'title':            _TD(N_('Title'),        'text'),
    'tracknumber':      _TD(N_('Track'),        'dblnum', min=0, max=500),
    'version':          _TD(N_('Version'),      'text'),
    'website':          _TD(N_('Website'),      'text'),

    # various internal tags

    '__bitrate':        _TD(N_('Bitrate'),      'bitrate', editable=False),
    '__basedir':        None,
    '__date_added':     _TD(N_('Date added'),   'timestamp', editable=False),
    '__last_played':    _TD(N_('Last played'),  'timestamp', editable=False),
    '__length':         _TD(N_('Length'),       'time', editable=False),
    '__loc':            _TD(N_('Location'),     'location', editable=False),
    '__modified':       _TD(N_('Modified'),     'timestamp', editable=False),
    '__playtime':       _TD(N_('Play time'),    'time', editable=False),
    '__playcount':      _TD(N_('Times played'), 'int', editable=False),
    '__rating':         None,  # currently special.
    '__startoffset':    _TD(N_('Start offset'), 'time', min=0, max=3600),  # TODO: calculate these parameters
    '__stopoffset':     _TD(N_('Stop offset'),  'time', min=0, max=3600),
    # fmt: on
}

disk_tags = set()

for k, v in tag_data.items():
    if v:
        v.tag_name = k
        if v.use_disk:
            disk_tags.add(k)


def get_default_tagdata(tag):
    """If the tagname is not in tag_data, you can use this function
    to get a _TD object for it"""

    return _TD(tag, 'text', editable=(not tag.startswith('__')), tag_name=tag)



# moodbar - Replace Exaile's seekbar with a moodbar
# Copyright (C) 2015  Johannes Sasongko <sasongko@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import hashlib
import os
from typing import Optional


class MoodbarCache:
    """Cache for moodbar files"""

    def __init__(self, location: str):
        try:
            os.mkdir(location)
        except OSError:
            pass
        self.__loc = location

    def get(self, uri: str) -> Optional[bytes]:
        try:
            with open(self._get_cache_path(uri), 'rb') as cachefile:
                return cachefile.read()
        except IOError:
            return None

    def put(self, uri: str, data: bytes) -> None:
        if data is None:
            return
        with open(self._get_cache_path(uri), 'wb') as cachefile:
            cachefile.write(data)

    def _get_cache_path(self, uri: str) -> str:
        hex_str = hashlib.sha256(uri.encode('utf-8')).hexdigest()
        return os.path.join(self.__loc, hex_str + '.mood')



# Copyright (C) 2011 Dustin Spicuzza
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.

#
# Grouping field utility functions
#


from gi.repository import Gtk
from gi.repository import GObject

import re

from xl import playlist, settings

from xl.nls import gettext as _
from xl.trax import search

from xlgui import main
from xlgui.widgets import dialogs

from . import gt_widgets

group_categories_option = 'plugin/grouptagger/group_categories'
migrated_option = 'plugin/grouptagger/0.2_migration'
tagname_option = 'plugin/grouptagger/tagname'


def migrate_settings():
    '''Automatically migrate group tagger 0.1 settings to 0.2'''

    if settings.get_option(migrated_option, False):
        default_groups = settings.get_option('plugin/grouptagger/default_groups', None)
        if default_groups is not None:
            group_categories = {_('Uncategorized'): [True, default_groups]}
            set_group_categories(group_categories)
            # settings.remove_option( 'plugin/grouptagger/default_groups' )

        settings.set_option(migrated_option, True)


def get_tagname():
    return settings.get_option(tagname_option, 'grouping')


def get_track_groups(track):
    """
    Returns a set() of groups present in this track
    """
    return _get_track_groups(track, get_tagname())


def _get_track_groups(track, tagname):
    grouping = track.get_tag_raw(tagname, True)

    if grouping is not None:
        return {group.replace('_', ' ') for group in grouping.split()}

    return set()


def set_track_groups(track, groups):
    """
    Given an array of groups, sets them on a track

    Returns true if successful, false if there was an error
    """

    grouping = ' '.join(sorted('_'.join(group.split()) for group in groups))
    track.set_tag_raw(get_tagname(), grouping)

    if not track.write_tags():
        dialogs.error(
            None,
            "Error writing tags to %s"
            % GObject.markup_escape_text(track.get_loc_for_io()),
        )
        return False

    return True


def get_group_categories():
    """
    Returns a dictionary that contains a mapping of default groups
    to categories.

    Structure: { category: [expanded, [group, ... ]], ... }
    """

    return settings.get_option(group_categories_option, dict())


def get_groups_from_categories():
    groups = set()
    categories = get_group_categories()
    for category, (expanded, cgroups) in categories.items():
        for group in cgroups:
            groups.add(group)
    return groups


def set_group_categories(group_categories):
    """
    Set the mapping of default groups to categories
    """
    settings.set_option(group_categories_option, group_categories)


def get_all_collection_groups(collection):
    """
    For a given collection of tracks, return all groups
    used within that collection
    """
    groups = set()
    for track in collection:
        groups |= get_track_groups(track)

    return groups


def _create_search_playlist(name, search_string, exaile):
    '''Create a playlist based on a search string'''
    tracks = [
        x.track
        for x in search.search_tracks_from_string(exaile.collection, search_string)
    ]

    # create the playlist
    pl = playlist.Playlist(name, tracks)
    main.get_playlist_notebook().create_tab_from_playlist(pl)


def create_all_search_playlist(groups, exaile):
    '''Create a playlist of tracks that have all groups selected'''

    tagname = get_tagname()

    name = '%s: %s' % (tagname.title(), ' and '.join(groups))
    search_string = ' '.join(
        [
            '%s~"\\b%s\\b"' % (tagname, re.escape(group.replace(' ', '_')))
            for group in groups
        ]
    )

    _create_search_playlist(name, search_string, exaile)


def create_custom_search_playlist(groups, exaile):
    '''Create a playlist based on groups, and user input in a shiny dialog'''

    dialog = gt_widgets.GroupTaggerQueryDialog(groups)
    if dialog.run() == Gtk.ResponseType.OK:
        name, search_string = dialog.get_search_params()
        _create_search_playlist(name, search_string, exaile)

    dialog.destroy()
